import logging as py_logging

from magma.config import config, EnvConfig
from magma.debug import debug_wire
from magma.logging import root_logger, StagedLogRecord
from magma.when import get_curr_block as get_curr_when_block


config._register(
    rewire_log_level=EnvConfig("MAGMA_REWIRE_LOG_LEVEL", "WARNING")
)


_logger = root_logger()


class WiringLogRecord(StagedLogRecord):
    def __init__(self, tpl: str, *values):
        super().__init__(tpl)
        self._values = values

    @staticmethod
    def _get_debug_name(value):
        if isinstance(value, int):
            return value
        return value.debug_name

    def args(self):
        return [
            WiringLogRecord._get_debug_name(value) for value in self._values
        ]


# TODO(rsetaluri): Remove this alias.
WiringLog = WiringLogRecord


class Wire:
    """
    Wire implements wiring.

    Each wire is represented by a bit.
    """

    def __init__(self, bit):
        self._bit = bit
        self._driving = []
        self._driver = None

    def __repr__(self):
        return repr(self._bit)

    def __str__(self):
        return str(self._bit)

    def anon(self):
        return self._bit.anon()

    def unwire(self, other=None, debug_info=None):
        if other is None:
            if self._driver is None:
                _logger.warning(
                    WiringLog("Unwire called on undriven value {}, ignoring",
                              self._bit),
                    debug_info=debug_info
                )
                return
            other = self._driver
        other._driving.remove(self)
        self._driver = None

    def connect(self, other, debug_info):
        """
        Connect two wires, self should be an input and other should be an
        output, or both should be inouts
        """
        if self._driver is not None:
            # TODO(rsetaluri): Issue a different warning in the case of the old
            # driver being a "conditional" driver (i.e. from a when context).
            old_driver = self._driver
            self.unwire(old_driver)
            _logger.log(
                py_logging.getLevelName(config.rewire_log_level),
                WiringLog(
                    ("Wiring multiple outputs to same wire, using last "
                     "connection. Input: {}, Old Output: {}, New Output: {}"),
                    self._bit, old_driver._bit, other._bit),
                debug_info=debug_info
            )
        if self._bit.is_output():
            _logger.error(
                WiringLog("Using `{}` (an output) as an input", self._bit),
                debug_info=debug_info
            )
            return
        if other._bit.is_input():
            _logger.error(
                WiringLog("Using `{}` (an input) as an output", other._bit),
                debug_info=debug_info
            )
            return
        if self._bit.is_inout() and not other._bit.is_inout():
            _logger.error(
                WiringLog("Using `{}` (not inout) as an inout", other._bit),
                debug_info=debug_info
            )
            return
        if not self._bit.is_inout() and other._bit.is_inout():
            _logger.error(
                WiringLog("Using `{}` (not inout) as an inout", self._bit),
                debug_info=debug_info
            )
            return

        self._driver = other
        other._driving.append(self)

    def trace(self, skip_self=True):
        """
        If a value is an input or an intermediate (undirected), trace it until
        there is an input or inout (this is the source)

        Upon the first invocation (from a user), we skip the current bit (so
        we don't trace to ourselves)
        """
        if self._driver is not None:
            return self._driver.trace(skip_self=False)
        if not skip_self and (self._bit.is_output() or self._bit.is_inout()):
            return self._bit
        if not skip_self and self._bit.has_children():
            # Could be an Array driven non-bulk, so check if children can be
            # traced
            return self._bit.trace(skip_self=False)
        return None

    def value(self):
        """
        Return the bit connected to this bit. Specifically, return the bit this
        bit is driving if it exists. Else if, there is a unique "drivee" bit,
        return that bit. Otherwise, return None.
        """
        if self._driver is not None:
            return self._driver._bit
        if len(self._driving) == 1:
            return self._driving[0]._bit
        return None

    def driven(self):
        return self._driver is not None

    def wired(self):
        return self._driver or self._driving

    def driving(self):
        """
        Return a (possibly empty) list of all bits this bit is driving.
        """
        return [driving._bit for driving in self._driving]

    @property
    def driver(self):
        return self._driver

    @property
    def bit(self):
        return self._bit


class Wireable:
    def __init__(self):
        self._wire = Wire(self)
        self._when_context = get_curr_when_block()

    def set_when_context(self, ctx):
        self._when_context = ctx

    def wired(self):
        return self._wire.wired()

    # return the input or output Bit connected to this Bit
    def trace(self, skip_self=True):
        return self._wire.trace(skip_self)

    # return the output Bit connected to this input Bit
    def value(self):
        return self._wire.value()

    def driven(self):
        return self._wire.driven()

    def driving(self):
        return self._wire.driving()

    def unwire(i, o=None, debug_info=None):
        if o is not None:
            o = o._wire
        i._wire.unwire(o, debug_info)

    def _wire_impl(self, o, debug_info):
        self._wire.connect(o._wire, debug_info)
        self.debug_info = debug_info
        o.debug_info = debug_info

    def wire(self, o, debug_info):
        curr_when_block = get_curr_when_block()
        is_conditional = (
            curr_when_block is not None
            and curr_when_block is not self._when_context
        )
        if not is_conditional:
            self._wire_impl(o, debug_info)
            return
        curr_when_block.add_conditional_wire(self, o)

    @debug_wire
    def rewire(self, o, debug_info=None):
        self.unwire(debug_info=debug_info)
        self.wire(o, debug_info=debug_info)



import abc
import dataclasses
import typing

import hwtypes as ht

from magma.array import Array
from magma.bit import Bit
from magma.bits import Bits, UInt, SInt
from magma.common import replace_all
from magma.digital import Digital
from magma.protocol_type import MagmaProtocol
from magma.t import Kind
from magma.tuple import Tuple, Product


def isprotocol(T):
    return issubclass(T, MagmaProtocol)


def isdigital(T):
    return issubclass(T, Digital)


def isbits(T):
    return issubclass(T, Bits)


def isarray(T):
    return issubclass(T, Array)


def istuple(T):
    return issubclass(T, Tuple)


def isuint(T):
    return issubclass(T, UInt)


def issint(T):
    return issubclass(T, SInt)


def _unwrap_protocol_type(T):
    if isprotocol(T):
        return T._to_magma_()
    return T


def _get_unbound_root(T):
    types = [T]
    while types:
        T = types.pop(0)
        if T._unbound_base_ is not None and not T._unbound_base_.is_bound:
            return T._unbound_base_
        types += T.__bases__


@dataclasses.dataclass(frozen=True)
class GenericWrapper:
    raw: typing.Any
    children: typing.List[typing.Any]
    name: str


class WrappedVisitor:
    def visit(self, node):
        wrapped = self.wrap(node)
        method = 'visit_' + wrapped.name
        visitor = getattr(self, method, self.generic_visit)
        return visitor(wrapped.raw)

    def generic_visit(self, node):
        wrapped = self.wrap(node)
        for child in wrapped.children:
            self.visit(child)

    @abc.abstractmethod
    def wrap(self, node):
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class _TypeWrapper(GenericWrapper):
    attrs: typing.Dict[str, typing.Any]
    constructor: typing.Callable


def _wrap_type(T):
    if isinstance(T, _TypeWrapper):
        return T
    T = _unwrap_protocol_type(T)

    if isdigital(T):
        return _TypeWrapper(T, [], "Digital", {}, lambda T: T)
    if isbits(T):
        return _TypeWrapper(T, [], "Bits", {}, lambda T: T)
    if isarray(T):

        def _constructor(T, subT):
            return T[T.N, subT]

        return _TypeWrapper(T, [T.T], "Array", {"N": T.N}, _constructor)
    if istuple(T):

        def _constructor(T, *args):
            field_dict = {key: arg
                          for key, arg in zip(T.field_dict.keys(), args)
                          if arg is not None}
            unbound = _get_unbound_root(T)
            return unbound.from_fields(T.__name__, field_dict)

        attrs = {"fields": T.field_dict}
        children = list(T.field_dict.values())
        return _TypeWrapper(T, children, "Tuple", attrs, _constructor)

    raise NotImplementedError(T)


class TypeVisitor(WrappedVisitor):
    def wrap(self, T):
        return _wrap_type(T)


class TypeTransformer(TypeVisitor):
    def generic_visit(self, T):
        wrapped = _wrap_type(T)
        args = (self.visit(child) for child in wrapped.children)
        return wrapped.constructor(T, *args)


def to_hwtypes(T: Kind):
    if not isinstance(T, Kind):
        raise TypeError("Expected a magma type")
    if issubclass(T, Bit):
        return ht.Bit
    if issubclass(T, SInt):
        return ht.SIntVector[T.N]
    if issubclass(T, UInt):
        return ht.UIntVector[T.N]
    if issubclass(T, Bits):
        return ht.BitVector[T.N]
    if issubclass(T, Product):
        _fields = {k: to_hwtypes(v) for k, v in T.field_dict.items()}
        return ht.Product.from_fields(T.__name__, _fields)
    if issubclass(T, Tuple):
        _fields = tuple(to_hwtypes(v) for v in T.fields)
        return ht.Tuple[_fields]
    raise NotImplementedError(T)


_TYPE_STR_SANITIZATION_MAP = {
    "(": "",
    ")": "",
    ",": "_",
    "=": "_",
    "[": "",
    "]": "",
    " ": "",
    "<": "",
    ">": "",
    ".": "",
    "'": "",
    ":": "",
    "{": "",
    "}": "",
}


def type_to_sanitized_string(T: Kind) -> str:
    return replace_all(str(T), _TYPE_STR_SANITIZATION_MAP)



import abc
import functools
import itertools
from typing import Dict, Iterable, Optional, Set, Tuple, Union

from magma.bits import Bits
from magma.circuit import Circuit, CircuitType, CircuitBuilder
from magma.conversions import from_bits
from magma.digital import Digital
from magma.primitives.register import Register
from magma.t import Type, In, Out
from magma.when import (
    BlockBase as WhenBlock,
    get_all_blocks as get_all_when_blocks,
    no_when,
    find_inferred_latches,
)
from magma.value_utils import make_selector
from magma.wire import wire


_ISWHEN_KEY = "_iswhen_"


def iswhen(inst_or_defn: Union[Circuit, CircuitType]) -> bool:
    if isinstance(inst_or_defn, Circuit):
        return getattr(type(inst_or_defn), _ISWHEN_KEY, False)
    return getattr(inst_or_defn, _ISWHEN_KEY, False)


class InferredLatchError(Exception):
    pass


# NOTE(rsetaluri): This is needed because importing magma.primitives.Memory
# results in a circular import.
@functools.lru_cache(maxsize=1)
def _Memory():
    from magma.primitives.memory import Memory
    return Memory


def _is_memory_port(value: Type) -> bool:
    Memory = _Memory()
    root_ref = value.name.root()
    try:
        value_inst = root_ref.inst
    except AttributeError:
        return False
    if not isinstance(type(value_inst), Memory):
        return False
    if root_ref.name not in ("WDATA", "WADDR", "WE", "RADDR"):
        return False
    return True


def _add_default_drivers_to_memory_ports(builder: 'WhenBuilder'):
    for drivee in builder.output_to_index:
        if drivee in builder.default_drivers:
            continue
        if not _is_memory_port(drivee):
            continue
        T = type(drivee).undirected_t
        zero = from_bits(T, Bits[T.flat_length()](0))
        builder.block.add_default_driver(drivee, zero)


def _get_corresponding_register_output(value: Type) -> Optional[Type]:
    root_ref = value.name.root()
    try:
        value_inst = root_ref.inst
    except AttributeError:
        return None
    if not isinstance(type(value_inst), Register):
        return None
    sel = make_selector(value)
    return sel.select(value_inst.O)


def _add_default_drivers_to_register_inputs(builder: 'WhenBuilder'):
    for drivee in builder.output_to_index:
        if drivee in builder.default_drivers:
            continue
        O = _get_corresponding_register_output(drivee)
        if O is None:
            continue
        builder.block.add_default_driver(drivee, O)


class WhenBuilder(CircuitBuilder):
    def __init__(self, block: WhenBlock):
        super().__init__(name=f"When_{id(block)}")
        self._block = block
        self._condition_counter = itertools.count()
        self._driver_counter = itertools.count()
        self._drivee_counter = itertools.count()
        self._input_counter = itertools.count()
        self._output_counter = itertools.count()
        self._input_to_index = {}
        self._output_to_index = {}
        self._default_drivers = {}
        self._set_definition_attr("primitive", True)
        self._set_definition_attr(_ISWHEN_KEY, True)
        self._set_definition_attr("_builder_", self)

    @property
    def default_drivers(self) -> Dict[Type, Type]:
        return self._default_drivers.copy()

    @property
    def input_to_index(self) -> Dict[Type, int]:
        return self._input_to_index.copy()

    @property
    def output_to_index(self) -> Dict[Type, int]:
        return self._output_to_index.copy()

    def _generic_add(
            self,
            value,
            value_to_index,
            index_counter,
            name_counter,
            name_prefix,
            type_qualifier,
    ):
        if value is None or value in value_to_index:
            return
        value_to_index[value] = next(index_counter)
        port_name = f"{name_prefix}{next(name_counter)}"
        self._add_port(port_name, type_qualifier(type(value).undirected_t))
        with no_when():
            wire(getattr(self, port_name), value)

    def add_drivee(self, value: Type):
        self._generic_add(
            value,
            self._output_to_index,
            self._output_counter,
            self._drivee_counter,
            "O",
            Out,
        )

    def add_driver(self, value: Type):
        self._generic_add(
            value,
            self._input_to_index,
            self._input_counter,
            self._driver_counter,
            "I",
            In,
        )

    def add_condition(self, condition: Optional[Digital]):
        self._generic_add(
            condition,
            self._input_to_index,
            self._input_counter,
            self._condition_counter,
            "S",
            In,
        )

    def add_default_driver(self, drivee: Type, driver: Type):
        self.add_driver(driver)
        self.add_drivee(drivee)
        self._default_drivers[drivee] = driver

    def _finalize(self):
        # NOTE(rsetaluri): These passes should ideally be done after circuit
        # creation. However, it is quite unwieldy, so we opt to do it in this
        # builder's finalization, although it is a bit "hacky". Ideally, this
        # builder should be completely agnostic of anything it is connected to,
        # and a global (post-processing) pass should be used instead.
        _add_default_drivers_to_memory_ports(self)
        _add_default_drivers_to_register_inputs(self)
        # Detect latches which would be inferred from the context of the when
        # block.
        latches = find_inferred_latches(self.block)
        if latches:
            raise InferredLatchError(latches)

    @property
    def block(self) -> WhenBlock:
        return self._block



"""
Definition of magma's Bit type
* Subtype of the Digital type
* Implementation of hwtypes.AbstractBit
"""
import keyword
import typing as tp
import functools
import hwtypes as ht
from hwtypes.bit_vector_abc import AbstractBit, TypeFamily
from .t import Direction
from .digital import Digital, DigitalMeta
from .digital import VCC, GND  # TODO(rsetaluri): only here for b.c.

from magma.compatibility import IntegerTypes
from magma.debug import debug_wire
from magma.family import get_family
from magma.interface import IO
from magma.language_utils import primitive_to_python
from magma.protocol_type import magma_type, MagmaProtocol
from magma.operator_utils import output_only


def bit_cast(fn: tp.Callable[['Bit', 'Bit'], 'Bit']) -> \
        tp.Callable[['Bit', tp.Union['Bit', bool]], 'Bit']:
    @functools.wraps(fn)
    def wrapped(self: 'Bit', other: tp.Union['Bit', bool]) -> 'Bit':
        if isinstance(other, Bit):
            return fn(self, other)
        try:
            other = Bit(other)
        except TypeError:
            return NotImplemented
        return fn(self, other)
    return wrapped


_IMPLICITLY_COERCED_ITE_TYPES = (int, ht.BitVector, ht.Bit)


class Bit(Digital, AbstractBit, metaclass=DigitalMeta):
    __hash__ = Digital.__hash__

    @staticmethod
    def get_family() -> TypeFamily:
        return get_family()

    def __init__(self, value=None, name=None):
        super().__init__(name=name)
        if value is None:
            self._value = None
        elif isinstance(value, Bit):
            self._value = value._value
        elif isinstance(value, bool):
            self._value = value
        elif isinstance(value, int):
            if value not in {0, 1}:
                raise ValueError('Bit must have value 0 or 1 not {}'.format(value))
            self._value = bool(value)
        elif hasattr(value, '__bool__'):
            self._value = bool(value)
        else:
            raise TypeError("Can't coerce {} to Bit".format(type(value)))

    @property
    def direction(self):
        return type(self).direction

    def __invert__(self):
        # CoreIR uses not instead of invert name
        return type(self).undirected_t._declare_unary_op("not")()(self)

    @bit_cast
    @output_only("Cannot use == on an input")
    def __eq__(self, other):
        if self is other:
            return True
        # CoreIR doesn't define an eq primitive for bits
        return ~(self ^ other)

    @bit_cast
    @output_only("Cannot use != on an input")
    def __ne__(self, other):
        # CoreIR doesn't define an ne primitive for bits
        return self ^ other

    @bit_cast
    def __and__(self, other):
        return type(self).undirected_t._declare_binary_op("and")()(self, other)

    @bit_cast
    def __or__(self, other):
        return type(self).undirected_t._declare_binary_op("or")()(self, other)

    @bit_cast
    def __xor__(self, other):
        return type(self).undirected_t._declare_binary_op("xor")()(self, other)

    def ite(self, t_branch, f_branch):
        if isinstance(t_branch, list) and isinstance(f_branch, list):
            if not len(t_branch) == len(f_branch):
                raise TypeError("Bit.ite of two lists expects the same length")
            return [self.ite(x, y) for x, y in zip(t_branch, f_branch)]

        if self.const():
            return t_branch if bool(self) else f_branch
        if isinstance(t_branch, tuple):
            return tuple(self.ite(t, f) for t, f in zip(t_branch, f_branch))
        # NOTE(rsetaluri): CoreIR flips t/f cases.
        # NOTE(rseatluri): self._mux is monkey patched in
        # magma/primitives/mux.py to avoid circular dependency.
        return self._mux([f_branch, t_branch], self)

    @debug_wire
    def wire(self, o, debug_info):
        # Cast to Bit here so we don't get a Digital instead
        if isinstance(o, (IntegerTypes, bool, ht.Bit)):
            o = Bit(o)
        return super().wire(o, debug_info)


BitIn = Bit[Direction.In]
BitOut = Bit[Direction.Out]
BitInOut = Bit[Direction.InOut]



import dataclasses
import functools
from typing import Callable, Iterable

from magma.circuit import CircuitKind, CircuitType, DefineCircuitKind
from magma.clock import ClockTypes
from magma.conversions import as_bits
from magma.value_utils import ValueVisitor
from magma.t import Type


@dataclasses.dataclass(frozen=True)
class _StubbifierResult:
    modified: bool
    descend: bool


Stubbifier = Callable[[Type], _StubbifierResult]


def zero_stubbifier(value: Type) -> _StubbifierResult:
    if isinstance(value, ClockTypes):
        return _StubbifierResult(modified=False, descend=False)
    if value.is_output():
        value.unused()
        return _StubbifierResult(modified=True, descend=False)
    if value.is_input():
        bits = as_bits(value)
        bits @= 0
        return _StubbifierResult(modified=True, descend=False)
    return _StubbifierResult(modified=False, descend=True)


class NoOverrideDrivenStubbifierFactory:
    def __init__(self, stubbifier: Stubbifier):
        self._stubbifier = stubbifier

    def __call__(self, value: Type) -> bool:
        if value.value() is not None:
            return _StubbifierResult(modified=False, descend=False)
        return self._stubbifier(value)


no_override_driven_zero_stubbifier = (
    NoOverrideDrivenStubbifierFactory(zero_stubbifier)
)


class _StubifyVisitor(ValueVisitor):
    def __init__(self, stubbifier: Stubbifier):
        self._stubbifier = stubbifier
        self._modified = False

    @property
    def modified(self):
        return self._modified

    def generic_visit(self, value: Type):
        result = self._stubbifier(value)
        if result.modified:
            self._modified = True
        if not result.descend:
            return
        super().generic_visit(value)


def _stubify_impl(
        ports: Iterable[Type],
        stubbifier: Stubbifier = no_override_driven_zero_stubbifier
) -> bool:
    modified = False
    for port in ports:
        visitor = _StubifyVisitor(stubbifier)
        visitor.visit(port)
        modified = modified or visitor.modified
    return modified


def _stub_open(cls):
    raise NotImplementedError("Can not call open() on a circuit stub")


@functools.singledispatch
def stubify(
        interface,
        stubbifier: Stubbifier = no_override_driven_zero_stubbifier
):
    _ = _stubify_impl(interface.ports.values(), stubbifier)


@stubify.register
def _(
        obj: CircuitKind,
        stubbifier: Stubbifier = no_override_driven_zero_stubbifier
):
    ckt = obj
    with ckt.open():
        modified = _stubify_impl(ckt.interface.ports.values(), stubbifier)
    if modified:
        # NOTE(rsetaluri): This is a hack because we don't have good handling of
        # isdefinition when the circuit is modified. We should be doing that
        # more principled-ly. See https://github.com/phanrahan/magma/issues/929.
        ckt._is_definition = True
    # Set ckt.open to be a function which raises a NotImplementedError. Note
    # that we *can't* do this in the class itself, since we need to call open()
    # to tie the outputs first (in stubify()). Afterwards, we can override the
    # method.
    setattr(ckt, "open", classmethod(_stub_open))


def circuit_stub(
        cls=None,
        *,
        stubbifier: Stubbifier = no_override_driven_zero_stubbifier
):
    """
    Inspired by https://pybit.es/decorator-optional-argument.html.

    Stub-ifys a circuit, based on optional parameter @stubbifier. Default
    behavior is driving all outputs to zero. All inputs will be marked as unused
    automatically as well.
    """
    if cls is None:
        return functools.partial(circuit_stub, stubbifier=stubbifier)
    stubify(cls, stubbifier)
    return cls


class _CircuitStubMeta(DefineCircuitKind):
    def __new__(metacls, name, bases, dct):
        cls = super().__new__(metacls, name, bases, dct)
        # Only call stubify() on user circuits (i.e. do not call on CircuitStub
        # base class).
        if not dct.get("_circuit_base_", False):
            stubify(cls, stubbifier=no_override_driven_zero_stubbifier)
        return cls


class CircuitStub(CircuitType, metaclass=_CircuitStubMeta):
    _circuit_base_ = True



import dataclasses
from typing import Any, Callable, Iterable, Mapping, Union

from magma.array import Array, ArrayMeta
from magma.backend.mlir.common import make_unique_name
from magma.backend.mlir.graph_lib import Graph
from magma.circuit import Circuit, DefineCircuitKind
from magma.common import replace_all, Stack, SimpleCounter
from magma.ref import Ref, ArrayRef, TupleRef
from magma.t import Kind, Type
from magma.tuple import TupleMeta, Tuple as m_Tuple


ModuleLike = Union[DefineCircuitKind, Circuit]


_VALUE_OR_TYPE_TO_STRING_REPLACEMENTS = {
    "(": "_",
    ")": "",
    "[": "_",
    "]": "",
    " ": "",
    ",": "_",
    "=": "_",
    ".": "_",
}


def contains_tuple(T: Kind):
    if isinstance(T, TupleMeta):
        return True
    if isinstance(T, ArrayMeta):
        return contains_tuple(T.T)
    return False


def value_or_type_to_string(value_or_type: Union[Type, Kind]):
    if isinstance(value_or_type, Type):
        s = value_or_type.name.qualifiedname("_")
    else:
        s = str(value_or_type)
    return replace_all(s, _VALUE_OR_TYPE_TO_STRING_REPLACEMENTS)


@dataclasses.dataclass(frozen=True)
class ValueWrapper:
    id: str = dataclasses.field(default_factory=make_unique_name, init=False)
    name: str
    T: Kind

    def is_input(self) -> bool:
        return self.T.is_input()

    def is_output(self) -> bool:
        return self.T.is_output()

    def is_mixed(self) -> bool:
        return self.T.is_mixed()

    def __iter__(self) -> Iterable['ValueWrapper']:
        if isinstance(T, TupleMeta):
            return (
                ValueWrapper(f"{value_wrapper.name}_{key}", TT)
                for key, TT in self.T.field_dict.items()
            )
        if isinstance(T, ArrayMeta):
            return (
                ValueWrapper(f"{value_wrapper.name}[{key}]", self.T.T)
                for index in range(T.N)
            )
        raise NotImplementedError(f"{self} is not iterable")


ValueOrValueWrapper = Union[Type, ValueWrapper]


def value_or_value_wrapper_to_tree(
        value_or_value_wrapper: ValueOrValueWrapper,
        flatten_all_tuples: bool = False) -> Graph:
    tree = Graph()
    state = Stack()
    index = SimpleCounter()

    def _visit_leaf(v):
        tree.add_node(v, index=index.next())
        if not state:
            return
        tree.add_edge(state.peek(), v)

    def _pre_descend(v):
        tree.add_node(v, index=index.next())
        if state:
            tree.add_edge(state.peek(), v)
        state.push(v)

    def _post_descend(v):
        assert state.pop() is v

    visit_value_or_value_wrapper_by_direction(
        value_or_value_wrapper,
        _visit_leaf,
        _visit_leaf,
        flatten_all_tuples=flatten_all_tuples,
        pre_descend=_pre_descend,
        post_descend=_post_descend,
    )

    return tree


def visit_value_or_value_wrapper_by_direction(
        value_or_value_wrapper: ValueOrValueWrapper,
        input_visitor: Callable[[Type], Any],
        output_visitor: Callable[[Type], Any],
        **kwargs):

    pre_descend = kwargs.get("pre_descend", lambda _: None)
    post_descend = kwargs.get("post_descend", lambda _: None)
    inout_visitor = kwargs.get("inout_visitor", None)

    def descend(v):
        if not isinstance(v, (m_Tuple, Array)):
            raise TypeError(value)
        pre_descend(v)
        for item in v:
            visit_value_or_value_wrapper_by_direction(
                item, input_visitor, output_visitor, **kwargs)
        post_descend(v)

    flatten_all_tuples = kwargs.get("flatten_all_tuples", False)

    if flatten_all_tuples and contains_tuple(type(value_or_value_wrapper)):
        return descend(value_or_value_wrapper)
    if value_or_value_wrapper.is_input():
        return input_visitor(value_or_value_wrapper)
    if value_or_value_wrapper.is_output():
        return output_visitor(value_or_value_wrapper)
    if value_or_value_wrapper.is_mixed():
        return descend(value_or_value_wrapper)
    if inout_visitor is not None:
        return inout_visitor(value_or_value_wrapper)

    raise TypeError(value_or_value_wrapper)


class InstanceWrapper:

    @dataclasses.dataclass(frozen=True)
    class _Interface:
        ports: Mapping[str, ValueWrapper]

    def __init__(
            self,
            name: str,
            ports: Mapping[str, Kind],
            attrs: Mapping[str, Any]):
        self._name = name
        ports = {name: ValueWrapper(name, T) for name, T in ports.items()}
        self._interface = InstanceWrapper._Interface(ports)
        self._attrs = attrs
        for name, value in ports.items():
            setattr(self, name, value)

    def __repr__(self) -> str:
        return f"InstanceWrapper({self.name})"

    @property
    def name(self) -> str:
        return self._name

    @property
    def interface(self) -> _Interface:
        return self._interface

    @property
    def attrs(self) -> Mapping[str, Any]:
        return self._attrs.copy()


def safe_root(ref: Ref) -> Ref:
    """Returns the root ref of @ref."""
    # TODO(rsetaluri): This should be able to return `ref.root()`, but depends
    # on #990.
    parent = ref
    if isinstance(ref, ArrayRef):
        parent = ref.array.name
    elif isinstance(ref, TupleRef):
        parent = ref.tuple.name
    if parent is ref:
        return ref
    return safe_root(parent)



from typing import Callable, List, Mapping, Union

from magma.backend.mlir.mlir import MlirValue
from magma.backend.mlir.printer_base import PrinterBase


MlirValueList = List[MlirValue]
MlirValueOrMlirValueList = Union[MlirValue, MlirValueList]


def get_name_fn(raw_names: bool) -> Callable[[MlirValue], str]:
    if raw_names:
        return lambda value: value.raw_name
    return lambda value: value.name


def _maybe_wrap_value_or_value_list(
        value_or_value_list: MlirValueOrMlirValueList) -> MlirValueList:
    if isinstance(value_or_value_list, MlirValue):
        return [value_or_value_list]
    return value_or_value_list


def print_names(
        value_or_value_list: MlirValueOrMlirValueList,
        printer: PrinterBase,
        raw_names: bool = False):
    value_list = _maybe_wrap_value_or_value_list(value_or_value_list)
    get_name = get_name_fn(raw_names)
    printer.print(", ".join(get_name(v) for v in value_list))


def print_types(
        value_or_value_list: MlirValueOrMlirValueList, printer: PrinterBase):
    value_list = _maybe_wrap_value_or_value_list(value_or_value_list)
    printer.print(", ".join(v.type.emit() for v in value_list))


def print_signature(
        value_or_value_list: MlirValueOrMlirValueList,
        printer: PrinterBase,
        raw_names: bool = False):
    value_list = _maybe_wrap_value_or_value_list(value_or_value_list)
    get_name = get_name_fn(raw_names)
    signatures = (f"{get_name(v)}: {v.type.emit()}" for v in value_list)
    printer.print(", ".join(signatures))


def print_attr_dict(attr_dict: Mapping, printer: PrinterBase):
    attr_dict_to_string = ", ".join(f"{k} = {v}" for k, v in attr_dict.items())
    printer.print(f"{{{attr_dict_to_string}}}")



from collections import OrderedDict
from magma.passes.clock import drive_undriven_clock_types_in_inst

def compileclocks(cls):
    for instance in cls.instances:
        drive_undriven_clock_types_in_inst(cls, instance)

def compile(main):

    if main.name == "":
        code = "AnonymousCircuit("
        # Special case anonymous circuits for now
        for name, value in main.interface.ports.items():
            code += "\"{}\", {},".format(name, value)
        code = code[:-1] + ")"  # Replace final "," with a )
        return code

    compileclocks(main)

    defn = main.find(OrderedDict())

    code = ''
    for k, v in defn.items():
         #print('compiling', k)
         code += repr(v) + '\n'
    return code



from magma.is_definition import isdefinition
from magma.passes.passes import EditDefinitionPass, pass_lambda


def _terminate_if_unwired_output(port):
    if port.is_mixed():
        # list comp so it doesn't short circuit
        terminated = [_terminate_if_unwired_output(p) for p in port]
        return any(terminated)
    if port.is_output() and not port.wired():
        port.unused()
        return True
    return False


def _terminate_unused(interface):
    terminated = False
    for port in interface.ports.values():
        terminated |= _terminate_if_unwired_output(port)
    return terminated


class TerminateUnusedPass(EditDefinitionPass):
    def edit(self, circuit):
        if _terminate_unused(circuit.interface):
            circuit._is_definition = True
        if not isdefinition(circuit):
            return
        for inst in circuit.instances:
            _terminate_unused(inst.interface)


terminate_unused = pass_lambda(TerminateUnusedPass)



import abc
import functools
from typing import Iterable, List, Optional, Tuple

from magma.array import Array
from magma.circuit import Circuit
from magma.ref import get_ref_defn, get_ref_inst
from magma.t import Type
from magma.tuple import Tuple as m_Tuple
from magma.type_utils import isdigital, isarray, istuple


InstanceCollection = List[Circuit]
Connection = Tuple[Type, Type]


def _get_driving(value: Type) -> Iterable[Connection]:
    T = type(value)
    if isdigital(T):
        yield from ((value, drivee) for drivee in value.driving())
        return
    if not (isarray(T) or istuple(T)):
        raise TypeError(value)
    for elt in value:
        yield from _get_driving(elt)


class _GrouperBase(abc.ABC):
    def __init__(self, instances: InstanceCollection):
        self._instances = instances
        self._run = False

    @abc.abstractmethod
    def _visit_input_connection(self, driver: Type, drivee: Type):
        raise NotImplementedError()

    @abc.abstractmethod
    def _visit_output_connection(self, driver: Type, drivee: Type):
        raise NotImplementedError()

    @abc.abstractmethod
    def _visit_undriven_port(self, port: Type):
        raise NotImplementedError()

    def run(self):
        if self._run:
            raise RuntimeError("Can run grouper at most once")
        self._run_impl()
        self._run = True

    def _is_external(self, value: Type) -> Optional[bool]:
        """
        Returns whether @value is external to this group, or None if
        undecidable.
        """
        defn = get_ref_defn(value.name)
        if defn is not None:
            return True
        inst = get_ref_inst(value.name)
        if inst is not None:
            return inst not in self._instances
        assert not value.name.bound()
        return None

    def _run_on_input(self, port: Type) -> Iterable[Connection]:
        driver = port.trace()
        if driver is None:
            self._visit_undriven_port(port)
            return
        if driver.const():
            return
        is_external = self._is_external(driver)
        if is_external is not None:
            if is_external:
                yield (driver, port)
            return
        if not isinstance(port, (m_Tuple, Array)):
            raise TypeError(port)
        for elt in port:
            yield from self._run_on_input(elt)

    def _run_on_output(self, port: Type) -> Iterable[Connection]:
        for driver, drivee in _get_driving(port):
            is_external = self._is_external(drivee)
            if is_external is not None:
                if is_external:
                    yield (driver, drivee)
                continue
            raise NotImplementedError(
                f"{port}: driving annonymous value from compile guard not "
                f"supported"
            )

    def _run_on_port(self, port: Type):
        if port.is_input():
            for driver, drivee in self._run_on_input(port):
                self._visit_input_connection(driver, drivee)
        elif port.is_output():
            for driver, drivee in self._run_on_output(port):
                self._visit_output_connection(driver, drivee)
        else:  # mixed type or InOut
            if not isinstance(port, (m_Tuple, Array)):
                raise TypeError(value)

            for elt in port:
                self._run_on_port(elt)

    def _run_on_instance(self, instance: Circuit):
        for port in instance.interface.ports.values():
            self._run_on_port(port)

    def _run_impl(self):
        for instance in self._instances:
            self._run_on_instance(instance)


GrouperBase = _GrouperBase



import hashlib

from magma.generator import Generator2
from magma.interface import IO
from magma.t import Kind, In, Out
from magma.type_utils import type_to_sanitized_string


def _hash_expr(expr: str):
    hasher = hashlib.shake_128()
    hasher.update(expr.encode())
    return hasher.hexdigest(8)


class _InlineVerilogExpression(Generator2):
    def __init__(self, expr: str, T: Kind):
        self.expr = expr
        self.io = IO(O=Out(T.undirected_t))
        self.primitive = True
        type_str = type_to_sanitized_string(T)
        self.name = f"InlineVerilogExpression_{type_str}_{_hash_expr(expr)}"


InlineVerilogExpression = _InlineVerilogExpression



from abc import ABC, abstractmethod
from itertools import chain
from .common import deprecated
from .compatibility import IntegerTypes, StringTypes
from .protocol_type import MagmaProtocolMeta
from .ref import InstRef, DefnRef, LazyDefnRef, LazyInstRef, NamedRef
from .t import Type, Kind, Direction
import weakref


__all__  = ['make_interface']
__all__ += ['InterfaceKind', 'Interface', 'AnonymousInterface']
__all__ += ['IO', 'SingletonInstanceIO']


def _flatten(l):
    """Flatten an iterable of iterables to list."""
    return list(chain(*l))


def _make_interface_name(decl):
    return f"Interface"


def _is_valid_port(port):
    return isinstance(port, (Kind, Type, MagmaProtocolMeta))


def _parse_decl(decl):
    """
    Parse argument declaration of the form:

        (name0, type0, name1, type1, ..., namen, typen)
    """
    if len(decl) % 2:
        raise ValueError(f"Expected even number of arguments, got {len(decl)}")

    names = decl[::2]
    ports = decl[1::2]
    # If name is empty, convert to the index.
    names = [name if name else i for i, name in enumerate(names)]
    # Check that all ports are given as instances of Kind or Type.
    if not all(_is_valid_port(port) for port in ports):
        raise ValueError(f"Expected kinds or types, got {ports}")

    return names, ports


def _make_ref(name, inst, defn):
    assert not (inst is not None and defn is not None)
    if inst:
        return InstRef(inst, name)
    if defn:
        return DefnRef(defn, name)
    return NamedRef(name)


def _make_port(typ, ref, flip):
    if flip:
        typ = typ.flip()
    if isinstance(typ, MagmaProtocolMeta):
        return typ._from_magma_value_(typ._to_magma_()(name=ref))
    return typ(name=ref)


def _rename_ports(ports, renamed_ports):
    for name, port in ports.items():
        if name not in renamed_ports:
            continue
        port.name.name = renamed_ports[name]


def _make_wire_str(driver, value, wired):
    """
    Emit the wiring string for `driver` and `value` for the `__repr__` of an
    interface.

    Used by `_make_wires`, `wired` is a set of previously wired values, so in
    cases of fan-out wires are not done more than once.

    Handles non-whole values (e.g. arrays/tuple constructed with values coming
    from multiple sources) by emitting the wiring of the child values.
    """
    if (value, driver) in wired:
        return ""
    wired.add((value, driver))
    if driver.has_children() and driver.name.anon():
        return "".join(_make_wire_str(d, v, wired)
                       for d, v in zip(driver, value))
    iname = value.name.qualifiedname()
    oname = driver.name.qualifiedname()
    return f"wire({oname}, {iname})\n"


def _make_wires(value, wired):
    """
    Make the wires used in the `__repr__` of an interface.

    Handles the recursive types with mixed direction (e.g. Tuples containing
    inputs and outputs)

    Handles anonymous values by skipping them (we don't include anonymous
    temporaries in the representation)

    Handles non-whole values (e.g. arrays/tuple constructed with values coming
    from multiple sources)

    Traces from `value` to the source driver, emitting the wiring of the
    temporaries along the way.
    """
    if value.is_output():
        return ""
    if value.is_mixed():
        return "".join(_make_wires(v, wired)
                       for v, _ in value.connection_iter())
    driver = value.value()
    while driver is not None and driver.is_driven_anon_temporary():
        driver = driver.value()
    if driver is None:
        return ""
    if driver.has_children() and driver.name.anon():
        return "".join(_make_wires(v, wired)
                       for v, _ in value.connection_iter())
    s = ""
    while driver is not None:
        if (driver, value) in wired:
            break
        s += _make_wire_str(driver, value, wired)
        if not driver.driven():
            break
        value = driver
        driver = driver.value()
    return s


def _dict_from_decl(decl):
    return dict(zip(decl[::2], decl[1::2]))


class InterfaceKind(Kind):
    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)
        cls.ports = _dict_from_decl(cls._decl)

    def items(cls):
        return cls.ports.items()

    def __iter__(cls):
        return iter(cls.ports)

    def args_to_str(cls):
        args = [f"\"{arg}\"" if i % 2 == 0 else str(arg)
                for i, arg in enumerate(cls._decl)]
        return ", ".join(args)

    def __str__(cls):
        return f"Interface({cls.args_to_str()})"

    def __repr__(cls):
        return str(cls)

    def __eq__(cls, rhs):
        if not isinstance(rhs, InterfaceKind):
            return False
        return cls._decl == rhs._decl

    def __ne__(cls, rhs):
        return super().__ne__(rhs)

    def __hash__(cls):
        return super().__hash__()


class _InterfaceBase(Type):
    """Abstract Base Class for Interface types"""
    def __str__(self):
        return str(type(self))

    def __repr__(self):
        s = ""
        wired = set()
        for name, value in self.ports.items():
            if value.is_output():
                continue
            s += _make_wires(value, wired)
        return s

    @classmethod
    def items(cls):
        return cls.ports.items()

    def __iter__(self):
        return iter(self.ports)

    def __len__(self):
        return len(self.ports.keys())

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.ports[key]
        if isinstance(key, int):
            return self.arguments()[key]
        if isinstance(key, slice):
            return [self[i] for i in range(*key.indices(len(self)))]
        raise ValueError(f"Expected key as str, int, or slice, got {key} "
                         f"({type(key)})")

    def arguments(self):
        """Return all the argument ports."""
        return list(self.ports.values())

    def is_input(self, port, include_clocks=False):
        return (port.is_input() and
                (not port.is_clock() or include_clocks))

    def inputs(self, include_clocks=False):
        """Return all the argument input ports."""
        return list(filter(lambda x: self.is_input(x, include_clocks),
                           self.ports.values()))

    def inputs_by_name(self, include_clocks=False):
        inputs = []
        for name, port in self.ports.items():
            if self.is_input(port):
                inputs.append(name)
        return inputs

    def outputs(self):
        """Return all the argument output ports."""
        return list(filter(lambda port: port.is_output(), self.ports.values()))

    def args(self):
        """Return all the arguments as name, port."""
        return _flatten(self.ports.items())

    def decl(self):
        """
        Return all the arguments as name, flip(port) (same as the declaration).
        """
        return _flatten([name, type(port).flip()]
                        for name, port in self.ports.items())

    def inputargs(self):
        """Return all the input arguments as name, port."""
        return _flatten([name, port]
                        for name, port in self.ports.items()
                        if port.is_input() and not port.is_clock())

    def outputargs(self):
        """Return all the output arguments as name, port."""
        return _flatten([name, port]
                        for name, port in self.ports.items()
                        if port.is_output())

    def clockargs(self):
        """Return all the clock arguments as name, port."""
        return _flatten([name, port]
                        for name, port in self.ports.items()
                        if port.is_clock())

    def clockargnames(self):
        """Return all the clock argument names."""
        return [name for name, port in self.ports.items()
                if port.is_clock()]

    def isclocked(self):
        """Return True if this interface has a Clock."""
        return any(port.is_clock()
                   for port in self.ports.values())


class Interface(_InterfaceBase):
    """Interface class"""
    def __init__(self, decl, renamed_ports={}):
        """
        Assumes the port instances are provided:
            e.g. Interface('I0', In(Bit)(), 'I1', In(Bit)(), 'O', Out(Bit)())
        """

        def _map_name(name):
            if name in renamed_ports:
                raise NotImplementedError("Port renaming not implemented")
            # Convert integer to str, e.g. 0 to "0".
            if isinstance(name, IntegerTypes):
                return str(name)
            return name

        names, types = _parse_decl(decl)
        names = map(_map_name, names)
        self.ports = dict(zip(names, types))

    def __str__(self):
        s = ", ".join(f"{k}: {v}" for k, v in self.ports.items())
        return f"Interface({s})"


class AnonymousInterface(Interface):
    """
    Anonymous interfaces can have ports that are not inputs/output, so we need
    ot see if they trace to an input or if they are wired to inputs

    TODO: Need tests for all these cases
    """
    def outputs(self):

        def _is_output(port):
            return (port.is_output() or port.trace() is None and
                    not port.wired() and not port.is_input() and
                    not port.is_inout())

        return list(filter(_is_output, self.ports.values()))

    def is_input(self, port, include_clocks=False):
        return (port.is_input() or port.trace() is None and
                port.wired() and not port.is_output() and not port.is_inout() and
                (not port.is_clock() or include_clocks))


class _DeclareInterface(_InterfaceBase):
    """
    _DeclareInterface class.

    First, an Interface is declared:
        Interface = DeclareInterface('I0', In(Bit), 'I1', In(Bit), 'O', Out(Bit))

    Then, the interface is instanced:
        interface = Interface()
    """
    def __init__(self, renamed_ports={}, *, inst=None, defn=None):
        assert not (inst is not None and defn is not None)
        cls = type(self)
        names, types = _parse_decl(cls._decl)
        refs = (_make_ref(name, inst, defn) for name in names)
        args = zip(names, zip(types, refs))
        flip = defn is not None
        self.ports = {n: _make_port(t, r, flip) for n, (t, r) in args}
        for port in self.ports.values():
            port.name._value = weakref.ref(port)
        _rename_ports(self.ports, renamed_ports)


class _DeclareSingletonInterface(_DeclareInterface):
    """_DeclareSingletonInterface class"""
    def __init__(self, renamed_ports={}, *, inst=None, defn=None):
        assert not (inst is not None and defn is not None)
        # This interface declaration has singleton semantics for its module
        # definition (not instances or further definition instancing).
        cls = type(self)
        if cls._initialized:
            super().__init__(renamed_ports, inst=inst, defn=defn)
            return
        assert defn is not None
        cls._io.bind(defn)  # bind IO to @defn
        self.ports = dict(cls._io.ports.items())
        _rename_ports(self.ports, renamed_ports)  # rename ports
        cls._initialized = True


class _DeclareSingletonInstanceInterface(_DeclareInterface):
    """_DeclareSingletonInterface class"""
    def __init__(self, renamed_ports={}, *, inst=None, defn=None):
        assert not (inst is not None and defn is not None)
        # This interface declaration has singleton semantics for its module
        # definition (not instances or further definition instancing).
        cls = type(self)

        if defn:
            if cls._initialized:
                super().__init__(renamed_ports, inst=inst, defn=defn)
                return
            cls._io.bind(defn)  # bind IO to @defn
            self.ports = dict(cls._io.ports.items())
            _rename_ports(self.ports, renamed_ports)  # rename ports
            cls._initialized = True
            return
        if inst:
            if cls._initialized_inst:
                super().__init__(renamed_ports, inst=inst, defn=defn)
                return
            cls._io.bind_inst(inst)  # bind IO to @inst
            self.ports = dict(cls._io.inst_ports.items())
            _rename_ports(self.ports, renamed_ports)  # rename ports
            cls._initialized_inst = True
            return
        super().__init__(renamed_ports, inst=inst, defn=defn)


def make_interface(*decl):
    """Interface factory function."""
    name = _make_interface_name(decl)
    dct = dict(_decl=decl)
    return InterfaceKind(name, (_DeclareInterface,), dct)


@deprecated(msg="DeclareInterface() is deprecated, use make_interface() "
            "instead")
def DeclareInterface(*decl):
    return make_interface(*decl)


class IOInterface(ABC):
    """Abstract base class for IO-like classes"""
    @property
    @abstractmethod
    def ports(self):
        raise NotImplementedError()

    @abstractmethod
    def bind(self, defn):
        raise NotImplementedError()

    @abstractmethod
    def decl(self):
        raise NotImplementedError()

    @abstractmethod
    def make_interface(self):
        raise NotImplementedError()

    @abstractmethod
    def __add__(self, other):
        raise NotImplementedError()

    def __iadd__(self, other):
        # __iadd__ is explicitly overriden to enforce that it is non-mutating.
        return self + other

    def flip(self) -> "IOInterface":
        raise NotImplementedError()


class IO(IOInterface):
    """
    Class for creating an interface bundle.

    @kwargs: ordered dict of {name: type}, ala decl.
    """
    # Note that because we require kwargs to be ordered, we have a strong
    # requirement here for >= python version 3.6. See
    # https://www.python.org/dev/peps/pep-0468/.
    def __init__(self, **kwargs):
        self._ports = {}
        self._decl = []
        self._bound = False
        for name, typ in kwargs.items():
            self.add(name, typ)
            self._decl.extend((name, typ))

    @property
    def ports(self):
        return self._ports.copy()

    def bind(self, defn):
        if self._bound:
            raise Exception("Can not bind IO multiple times")
        for port in self._ports.values():
            port.name.set_defn(defn)
        self._bound = True

    def decl(self):
        return self._decl

    def make_interface(self):
        decl = self.decl()
        name = _make_interface_name(decl)
        dct = dict(_io=self, _decl=decl, _initialized=False)
        return InterfaceKind(name, (_DeclareSingletonInterface,), dct)

    def __add__(self, other):
        """
        Attempts to combine this IO and @other. Returns a new IO object with the
        combined ports, unless:
          * @other is not of type IOInterface, in which case a TypeError is
            raised
          * this or @other has already been bound, in which case an Exception is
            raised
          * this and @other have common port names, in which case an Exception
            is raised
        """
        if not isinstance(other, IOInterface):
            raise TypeError(f"unsupported operand type(s) for +: 'IO' and "
                            f"'{type(other).__name__}'")
        if self._bound or other._bound:
            raise Exception("Adding bound IO not allowed")
        if self._ports.keys() & other._ports.keys():
            raise Exception("Adding IO with duplicate port names not allowed")
        decl = self._decl + other._decl
        return IO(**_dict_from_decl(decl))

    def add(self, name, typ):
        if self._bound:
            raise RuntimeError("Can not add to a bound IO")
        # Definition port.
        ref = LazyDefnRef(name=name)
        port = _make_port(typ, ref, flip=True)
        self._ports[name] = port

    def _get(self, key: str):
        return self._ports[key]

    def __getitem__(self, key: str):
        return self._get(key)

    def __getattr__(self, key: str):
        try:
            return self._get(key)
        except KeyError:
            pass
        return super().__getattribute__(key)

    def fields(self):
        return _dict_from_decl(self._decl)

    def flip(self):
        return IO(**{name: T.flip() for name, T in self.fields().items()})


class SingletonInstanceIO(IO):
    """
    Class for creating an interface bundle for a singleton instance.

    @kwargs: ordered dict of {name: type}, ala decl.
    """
    def __init__(self):
        super().__init__()
        self._inst_ports = {}
        self._bound_inst = False

    def bind_inst(self, inst):
        if self._bound_inst:
            raise Exception("Can not bind IO multiple times")
        for port in self._inst_ports.values():
            port.name.set_inst(inst)
        self._bound = True

    @property
    def inst_ports(self):
        return self._inst_ports.copy()

    def decl(self):
        return _flatten((name, type(port))
                        for name, port in self._ports.items())

    def make_interface(self):
        decl = self.decl()
        name = _make_interface_name(decl)
        dct = dict(_io=self, _decl=decl, _initialized=False,
                   _initialized_inst=False)
        return InterfaceKind(name, (_DeclareSingletonInstanceInterface,), dct)

    def add(self, name, typ):
        super().add(name, typ)
        # Instance port.
        inst_ref = LazyInstRef(name=name)
        inst_port = _make_port(typ, inst_ref, flip=False)
        self._inst_ports[name] = inst_port

    def __add__(self, other):
        raise NotImplementedError(f"Addition operator disallowed on "
                                  f"{cls.__name__}")

    def flip(self):
        raise NotImplementedError()



import abc
import colorlog
import contextlib
import inspect
import io
import logging
import sys
import traceback

from magma.backend.util import make_relative
from magma.common import Stack
from magma.config import config, EnvConfig


config._register(
    log_stream=EnvConfig("MAGMA_LOG_STREAM", "stderr"),
    log_level=EnvConfig("MAGMA_LOG_LEVEL", "INFO"),
    include_traceback=EnvConfig("MAGMA_INCLUDE_WIRE_TRACEBACK", False, bool),
    traceback_limit=EnvConfig("MAGMA_ERROR_TRACEBACK_LIMIT", 5, int),
)


_staged_logs_stack = Stack()
_log_capturer_stack = Stack()


def _make_bold(string):
    return f"\033[1m{string}\033[0m"


def _get_source_line(filename, lineno):
    with open(filename, "r") as f:
        return f.readlines()[lineno - 1]


def _attach_debug_info(msg, debug_info):
    file = debug_info[0]
    line = debug_info[1]
    line_info = _make_bold(f"{make_relative(file)}:{line}")
    msg = f"{line_info}: {msg}"
    try:
        source = _get_source_line(file, line).rstrip("\n")
        source = f">> {source}"
    except FileNotFoundError:
        source = f"(Could not file file {file})"
    msg = f"{msg}\n{source}"
    return msg


def _attach_traceback(msg, frame_selector, limit):
    """
    Attaches traceback string to @msg and returns new string.

    @frame_selector is a function which takes a list of stack frames and selects
    one. For example, it could select the frame based on an index, or based on
    the function names.
    """
    frame = frame_selector(inspect.stack()).frame
    with io.StringIO() as io_:
        traceback.print_stack(f=frame, limit=limit, file=io_)
        tb = io_.getvalue()
    msg = f"{msg}\n{tb}"
    return msg


def _frame_selector(frames):
    return frames[3]


def _get_additional_kwarg(kwargs, key):
    try:
        value = kwargs.pop(key)
        return value
    except KeyError:
        return None


def get_staged_logs_stack() -> Stack:
    global _staged_logs_stack
    return _staged_logs_stack


class _MagmaLogger(logging.Logger):
    """
    Derivative of logging.Logger class, with two additional keyword args:
    * 'debug_info': Tuple of (file_name, line_no). If 'debug_info' is included,
       this source-level information is logged along with the message.
    * 'include_traceback': If True, a traceback is printed along with the
       message.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._raw = False

    @property
    def raw(self) -> bool:
        return self._raw

    @raw.setter
    def raw(self, raw: bool):
        self._raw = raw

    @contextlib.contextmanager
    def as_raw(self):
        prev_raw = self.raw
        self.raw = True
        try:
            yield self
        finally:
            self.raw = prev_raw

    def _log(self, level, msg, args, **kwargs):
        if not self.raw and self._staged_log(level, msg, args, **kwargs):
            return
        self._raw_log(level, msg, args, **kwargs)

    def _staged_log(self, level, msg, args, **kwargs) -> bool:
        staged_logs_stack = get_staged_logs_stack()
        try:
            staged_logs = staged_logs_stack.peek()
        except IndexError:
            return False
        staged_logs.append((self, level, msg, args, kwargs))
        return True

    def _capture_log(self, level, msg, args, **kwargs):
        try:
            log_capturer = get_log_capturer()
        except IndexError:
            return
        log_capturer.add_log((level, msg, args, kwargs))

    def _raw_log(self, level, msg, args, **kwargs):
        debug_info = _get_additional_kwarg(kwargs, "debug_info")
        if debug_info:
            msg = _attach_debug_info(msg, debug_info)
        include_traceback = _get_additional_kwarg(kwargs, "include_traceback")
        if include_traceback or config.include_traceback:
            msg = _attach_traceback(
                msg, _frame_selector, config.traceback_limit)
        self._capture_log(level, msg, args, **kwargs)
        super()._log(level, msg, args, **kwargs)


# Set logging class to _MagmaLogger to override logging behavior. Also, setup
# root logger parameters.
logging.setLoggerClass(_MagmaLogger)
_log_stream = getattr(sys, config.log_stream)
_root_logger = logging.getLogger("magma")
_handler = colorlog.StreamHandler(_log_stream)
_handler.setFormatter(colorlog.ColoredFormatter(
    '%(log_color)s%(levelname)s%(reset)s:%(name)s:%(message)s'))
_root_logger.addHandler(_handler)
_root_logger.setLevel(config.log_level)


def root_logger():
    return logging.getLogger("magma")


# NOTE(rsetaluri): For some reason the following code which uses
# contextlib.contextmanager results in the context manager being entered into
# twice. It may be cached somewhere in the pipeline.
#
#    @contextlib.contextmanager
#    def logging_level(level):
#        root = root_logger()
#        prev_level = root.level
#        root.setLevel(level)
#        try:
#            yield
#        finally:
#            root.setLevel(prev_level)
class logging_level:
    def __init__(self, level):
        self.level = level
        self.root = root_logger()

    def __enter__(self):
        self.prev_level = self.root.level
        self.root.setLevel(self.level)

    def __exit__(self, *_):
        self.root.setLevel(self.prev_level)


def stage_logger():
    get_staged_logs_stack().push([])


def _flush(staged_logs):
    for logger, level, obj, args, kwargs in staged_logs:
        with logger.as_raw():
            logger.log(level, obj, *args, **kwargs)


def flush():
    staged_logs = get_staged_logs_stack().pop()
    _flush(staged_logs)
    return staged_logs


def unstage_logger():
    return flush()


def flush_all():
    staged_logs_stack = get_staged_logs_stack()
    while staged_logs_stack:
        staged_logs = staged_logs_stack.pop()
        _flush(staged_logs)


@contextlib.contextmanager
def staged_logs():
    stage_logger()
    staged_logs = get_staged_logs_stack().peek()
    try:
        yield staged_logs
    finally:
        unstage_logger()


class StagedLogRecord(abc.ABC):
    def __init__(self, tpl: str):
        self._tpl = tpl

    @abc.abstractmethod
    def args(self):
        raise NotImplementedError()

    def __str__(self):
        return self._tpl.format(*self.args())


def _get_log_capturer_stack() -> Stack:
    global _log_capturer_stack
    return _log_capturer_stack


def push_log_capturer(log_capturer):
    _get_log_capturer_stack().push(log_capturer)


def pop_log_capturer():
    _get_log_capturer_stack().pop()


def get_log_capturer():
    return _get_log_capturer_stack().peek()


@contextlib.contextmanager
def capture_logs(log_capturer):
    push_log_capturer(log_capturer)
    try:
        yield
    finally:
        pop_log_capturer()



import contextlib
import hashlib
import string
from typing import Mapping, Optional, Union

from ast_tools.stack import _SKIP_FRAME_DEBUG_STMT, get_symbol_table

from magma.array import Array
from magma.bit import Bit
from magma.circuit import Circuit
from magma.clock import ClockTypes
from magma.definition_context import DefinitionContext, get_definition_context
from magma.digital import Digital
from magma.interface import IO
from magma.passes.passes import CircuitPass
from magma.primitives.wire import Wire
from magma.ref import DefnRef, InstRef, ArrayRef, TupleRef
from magma.t import Type, Direction, In
from magma.tuple import Tuple
from magma.view import PortView, InstView
from magma.wire import wire


ValueLike = Union[Type, PortView]


class InlineVerilogError(RuntimeError):
    pass


def _get_view_inst_parent(view):
    while not isinstance(view, InstView):
        assert isinstance(view, PortView), type(view)
        return _get_view_inst_parent(view.parent)
    return view


def _make_inline_value(
        inline_value_map: Mapping[str, ValueLike], value: ValueLike) -> str:
    if isinstance(value, Array) and not issubclass(value.T, Digital):
        key = ", ".join(
            _make_inline_value(inline_value_map, t)
            for t in reversed(value)
        )
        return f"'{{{key}}}"
    if isinstance(value, Tuple):
        raise NotImplementedError(value)
    key = f"__magma_inline_value_{len(inline_value_map)}"
    inline_value_map[key] = value
    return f"{{{key}}}"


def _make_temporary(
        defn,
        value: Type,
        idx: int,
        inline_wire_prefix: str,
        parent: Optional[InstView] = None) -> ValueLike:

    def _insert_wire():
        temp_name = f"{inline_wire_prefix}{idx}"
        temp = Wire(type(value).undirected_t)(name=temp_name)
        temp.I @= value
        return temp

    # Insert a wire so that the value can't be inlined. If @defn is not None
    # then it needs to be `open()`'ed.
    ctx = contextlib.nullcontext() if defn is None else defn.open()
    with ctx:
        temp = _insert_wire()
    if parent is not None:
        return PortView(temp.O, parent)
    return temp.O


def _insert_temporary_wires(
        context: DefinitionContext,
        value: ValueLike,
        inline_wire_prefix: str):
    """
    Insert a temporary Wire instance so the signal isn't inlined out.

    We have to do this for DefnRef because the coreir inline.cpp logic
    sometimes inserts temporary wires for DefnRef that eventually get inlined.
    """
    wire_map = context.set_default_metadata("inline_verilog_wire_map", {})
    if isinstance(value, Type):
        if value.is_input():
            orig_value = value
            value = value.value()
            if value is None:
                raise InlineVerilogError(
                    f"Found reference to undriven input port: "
                    f"{repr(orig_value)}")

        key = value
        is_undriven_clock = (isinstance(value, ClockTypes) and value.is_input()
                             and not value.driven())
        if is_undriven_clock:
            # Share wire for undriven clocks so we don't
            # generate a separate wire for the eventual
            # driver from the automatic clock wiring logic
            key = type(value)
        if key not in wire_map:
            temp = _make_temporary(None, value,
                                   len(wire_map),
                                   inline_wire_prefix)
            wire_map[key] = temp
        value = wire_map[key]
    else:
        assert isinstance(value, PortView)
        if value.port.is_input():
            # For now we can't handle input port views, since this might return
            # a connection to the interface (e.g. self.I), then we'd need to
            # walk up to the parent instance and select the corresponding input
            # port, then trace that, which may need to happen recursively until
            # we find the driver in the current definition.
            # We don't have any existing use case for this, and a work around
            # exists which requires the user to effectively use the driver we
            # would find, rather than selecting the hierarchical input port
            # (this should work for simple cases, but might be annoying for
            # deep hierarchies where the driver could come from a far away
            # place, at which point we can add this feature)
            raise NotImplementedError()
        if value not in wire_map:
            # get first instance parent, then the parent of that will
            # be the container where we insert a wire
            parent = _get_view_inst_parent(value).parent
            if isinstance(parent, InstView):
                defn = parent.circuit
            else:
                assert isinstance(parent, Circuit)
                defn = type(parent)
            temp = _make_temporary(defn, value.port,
                                   len(wire_map),
                                   inline_wire_prefix, parent)
            wire_map[value] = temp
        value = wire_map[value]
    return value


def _build_io(inline_value_map: Mapping[str, ValueLike]) -> IO:
    if not inline_value_map:
        # Add dummy port so IO is not empty.
        io = IO(I=In(Bit))
        io.I.unused()
        return io
    io = IO()
    for key, value in inline_value_map.items():
        if isinstance(value, PortView):
            T = value.T
        else:
            T = type(value)
        io += IO(**{key: In(T)})
    return io


def _inline_verilog(
        context: DefinitionContext,
        format_str: str,
        format_args: Mapping[str, ValueLike],
        inline_value_map: Mapping[str, ValueLike],
        inline_wire_prefix: str):
    inline_verilog_modules = context.set_default_metadata(
        "inline_verilog_modules", [])
    format_args = format_args.copy()
    for key, arg in format_args.items():
        if isinstance(arg, (Type, PortView)):
            arg = _make_inline_value(inline_value_map, arg)
        format_args[key] = arg
    inline_str = format_str.format(**format_args)

    # Because modules/instances are sorted lexigraphically in the generated
    # verilog, in order for the inline verilog in the output to be in
    # statement-order, we need to generate a lexicographically increasing string
    # based on the order of the inline verilog statement in the python. A trick
    # to do this with numbers is, for every 10 modules, to insert a prefix 9 so
    # that it comes after the previous 1-9 digits.
    i = len(inline_verilog_modules)
    prefix = "9" * (i // 10)
    suffix = i % 10

    class _InlineVerilog(Circuit):
        name = f"{context.placer.name}_inline_verilog_{prefix}{suffix}"
        io = _build_io(inline_value_map)
        # Since each inline value (key) maps to a port, we populate the
        # connect_references dictionary with a map from key to port.
        connect_references = {}
        for key in inline_value_map:
            port = getattr(io, key)
            port.unused()  # these are needed so CoreIR knows it's a definition
            connect_references[key] = port
        inline_verilog_strs = [(inline_str, connect_references)]

    inline_verilog_modules.append(_InlineVerilog)
    inst_name = f"{context.placer.name}_inline_verilog_inst_{prefix}{suffix}"
    inst = _InlineVerilog(name=inst_name)
    # If there no interpolated values, _build_io adds a dummy port to avoid an
    # empty IO. We drive that with 0 here to avoid a hanging input.
    if not inline_value_map:
        inst.I @= 0

    for key, value in inline_value_map.items():
        value = _insert_temporary_wires(context, value, inline_wire_prefix)
        wire(value, getattr(inst, key))


def _process_fstring_syntax(
        format_str: str,
        format_args: Mapping[str, ValueLike],
        inline_value_map: Mapping[str, ValueLike],
        symbol_table: Mapping) -> str:
    fieldnames = [
        fname
        for _, fname, _, _ in string.Formatter().parse(format_str)
        if fname
    ]
    for field in fieldnames:
        if field in format_args:
            continue
        try:
            value = eval(field, {}, symbol_table)
        except NameError:
            continue
        if isinstance(value, (Type, PortView)):
            # Magma values (more precisely value-like objects, i.e. values or
            # PortView objects) are interpolated into strings specially. We
            # handle that here by creating a wire instance to make sure the
            # values persist.
            value = _make_inline_value(inline_value_map, value)
            # Stage for subsequent format call for regular kwargs inside
            # _inline_verilog.
            value = value.replace("{", "{{").replace("}", "}}")
        format_str = format_str.replace(f"{{{field}}}", str(value))
    return format_str


def _process_inline_verilog(
        context: DefinitionContext,
        format_str: str,
        format_args: Mapping[str, ValueLike],
        symbol_table: Mapping,
        inline_wire_prefix: str):
    inline_value_map = {}
    if symbol_table is not None:
        format_str = _process_fstring_syntax(
            format_str, format_args, inline_value_map, symbol_table)
    _inline_verilog(
        context, format_str, format_args, inline_value_map, inline_wire_prefix)


def inline_verilog_impl(
        format_str: str,
        format_args: Mapping[str, ValueLike],
        symbol_table: Mapping,
        inline_wire_prefix: str):
    context = get_definition_context()
    _process_inline_verilog(
        context, format_str, format_args, symbol_table, inline_wire_prefix)


def inline_verilog(
        format_str: str,
        inline_wire_prefix: str = "_magma_inline_wire",
        **kwargs):
    exec(_SKIP_FRAME_DEBUG_STMT)
    format_args = kwargs
    symbol_table = get_symbol_table([inline_verilog], copy_locals=True)
    inline_verilog_impl(
        format_str, format_args, symbol_table, inline_wire_prefix)


class ProcessInlineVerilogPass(CircuitPass):
    def __call__(self, _):
        pass



from typing import Dict, Optional

from magma.circuit import get_instance_callback
from magma.passes.passes import InstancePass


class InstanceCallbackPass(InstancePass):
    def __init__(self, main, data: Optional[Dict]):
        super().__init__(main)
        if data is None:
            data = {}
        self.data = data

    def __call__(self, path):
        inst = path[-1]
        try:
            callback = get_instance_callback(inst)
        except AttributeError:
            return
        path = [self.main] + list(path)
        callback(path, self.data)


def instance_callback_pass(main, data: Optional[Dict] = None):
    p = InstanceCallbackPass(main, data)
    p.run()
    return p.data



import collections
import dataclasses
import functools
from typing import Any, Callable, Dict, Iterable, Optional, Tuple


def wrap_with_not_implemented_error(fn):

    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        ret = fn(*args, **kwargs)
        if ret is None:
            raise NotImplementedError(args, kwargs)
        return ret

    return wrapped


_unique_name_index = 0


def make_unique_name() -> str:
    global _unique_name_index
    name = "%032d" % _unique_name_index
    _unique_name_index += 1
    return name


def default_field(cons, **kwargs):
    return dataclasses.field(default_factory=cons, **kwargs)


@dataclasses.dataclass
class WithId:
    id: str = dataclasses.field(default_factory=make_unique_name, init=False)


def constant(value: Any):
    return lambda: value


def try_call(fn: Callable[[], Any], ExceptionType: Optional[Any]):
    if ExceptionType is None:
        ExceptionType = BaseException
    try:
        ret = fn()
    except ExceptionType:
        pass
    else:
        return ret



"""
Based on chisel BitPat
https://github.com/freechipsproject/chisel3/blob/v3.3.2/src/main/scala/chisel3/util/BitPat.scala
"""
import string

from hwtypes import BitVector

from magma.bits import Bits


class BitPattern:
    def __init__(self, pattern):
        """
        Parses a bit pattern string `pattern` into the attributes `bits`,
        `mask`, `width`

        * `bits`  - the literal value, with don't cares being 0
        * `mask`  - the mask bits, with don't cares being 0 and cares being 1
        * `width` - the number of bits in the literal, including values and
                    don't cares, but not including the white space and
                    underscores
        """
        if pattern[0] != "b":
            raise ValueError("BitPattern must be in binary and prefixed with "
                             "'b'")
        bits = 0
        mask = 0
        count = 0
        for digit in pattern[1:]:
            if digit == '_' or digit in string.whitespace:
                continue
            if digit not in "01?":
                raise ValueError(
                    f"BitPattern {pattern} contains illegal character: {d}"
                )
            mask = (mask << 1) + (0 if digit == "?" else 1)
            bits = (bits << 1) + (1 if digit == "1" else 0)
            count += 1
        self.bits = BitVector[count](bits)
        self.mask = BitVector[count](mask)
        self.width = count
        self.const = self.mask == ((1 << self.width) - 1)
        self.__hash = hash(pattern)

    def __eq__(self, other):
        if not isinstance(other, Bits):
            raise TypeError(
                "BitPattern can only be compared to Bits"
            )
        return self.bits == (other & self.mask)

    def as_bv(self):
        if not self.const:
            raise TypeError(
                "Can only convert BitPattern with no don't cares to int"
            )
        return self.bits

    def __hash__(self):
        return self.__hash



from typing import Union

from magma.array import ArrayMeta
from magma.backend.mlir.magma_common import (
    InstanceWrapper, value_or_type_to_string)
from magma.bits import BitsMeta
from magma.digital import DigitalMeta
from magma.generator import Generator2
from magma.interface import IO
from magma.t import In, Out
from magma.tuple import TupleMeta, ProductMeta


def _get_tuple_field_type(T: TupleMeta, index: Union[int, str]):
    if isinstance(T, ProductMeta):
        return T.field_dict[index]
    index = int(index)
    return T.field_dict[index]


def _resized_array(T: ArrayMeta, new_size: int):
    if isinstance(T, BitsMeta):
        return T.unsized_t[new_size]
    return T[new_size, T.T]


def MagmaArrayGetOp(T: ArrayMeta, index: int):
    assert isinstance(T, ArrayMeta)
    T = T.undirected_t
    name = f"magma_array_get_op_{value_or_type_to_string(T)}"
    T_out = T.T
    ports = dict(I=In(T), O=Out(T_out))
    attrs = dict(T=T, index=index)
    return InstanceWrapper(name, ports, attrs)


def MagmaArraySliceOp(T: ArrayMeta, lo: int, hi: int):
    assert isinstance(T, ArrayMeta)
    T = T.undirected_t
    type_string = value_or_type_to_string(T)
    name = f"magma_array_slice_op_{type_string}_{lo}_{hi}"
    T_out = _resized_array(T, hi - lo)
    ports = dict(I=In(T), O=Out(T_out))
    attrs = dict(T=T, lo=lo, hi=hi)
    return InstanceWrapper(name, ports, attrs)


def MagmaArrayCreateOp(T: ArrayMeta):
    assert isinstance(T, ArrayMeta)
    T = T.undirected_t
    name = f"magma_array_create_op_{value_or_type_to_string(T)}"
    ports = {f"I{i}": In(T.T) for i in range(T.N)}
    ports.update(dict(O=Out(T)))
    attrs = dict(T=T)
    return InstanceWrapper(name, ports, attrs)


def MagmaTupleGetOp(T: TupleMeta, index: Union[int, str]):
    assert isinstance(T, TupleMeta)
    T = T.undirected_t
    name = f"magma_tuple_get_op_{value_or_type_to_string(T)}_{index}"
    T_out = _get_tuple_field_type(T, index)
    ports = dict(I=In(T), O=Out(T_out))
    attrs = dict(T=T, index=index)
    return InstanceWrapper(name, ports, attrs)


def MagmaTupleCreateOp(T: TupleMeta):
    assert isinstance(T, TupleMeta)
    T = T.undirected_t
    name = f"magma_tuple_create_op_{value_or_type_to_string(T)}"
    fields = T.field_dict
    ports = {f"I{k}": In(t) for k, t in fields.items()}
    ports.update(dict(O=Out(T)))
    return InstanceWrapper(name, ports, {})


def MagmaBitConstantOp(T: DigitalMeta, value: bool):
    assert isinstance(T, DigitalMeta)
    T = T.undirected_t
    name = f"magma_bit_constant_op_{value_or_type_to_string(T)}"
    ports = dict(O=Out(T))
    attrs = dict(value=value)
    return InstanceWrapper(name, ports, attrs)


def MagmaBitsConstantOp(T: BitsMeta, value: int):
    assert isinstance(T, BitsMeta)
    T = T.undirected_t
    name = f"magma_bits_constant_op_{value_or_type_to_string(T)}"
    ports = dict(O=Out(T))
    attrs = dict(value=value)
    return InstanceWrapper(name, ports, attrs)



import gc
from .passes import DefinitionPass

__all__ = ['DebugNamePass']


class InstDecl:
    def __init__(self, varname, lineno, filename):
        self.varname = varname
        self.lineno = lineno
        self.filename = filename

    def __repr__(self):
        return f"{self.filename} line {self.lineno}: {self.varname}"


def _create_debugname(ref_name, ref, inst):
    if ref is inst:
        return ref_name
    if isinstance(ref, list):
        idx = ref.index(inst)
        return f"{ref_name}[{idx}]"
    return ref_name


class DebugNamePass(DefinitionPass):
    def __call__(self, definition):
        for inst in definition.instances:
            referrers = gc.get_referrers(inst)
            # Skip if this pass has already been run (partially or not).
            if hasattr(inst, "decl"):
                continue
            stack = inst.stack
            inst.decl = None
            # Skip the first element in the stack, where the inst is placed.
            for frame_info in stack[1:]:
                local_vars = frame_info[0].f_locals.items()
                for name, var in local_vars:
                    if any(var is x for x in referrers) or var is inst:
                        debugname = _create_debugname(name, var, inst)
                        # Ignore the 'self' frame (where the inst is created).
                        if debugname == "self":
                            break
                        lineno = frame_info[2]
                        filename = frame_info[0].f_code.co_filename
                        inst.decl = InstDecl(debugname, lineno, filename)
                        break



from magma.generator import Generator2
from magma.interface import IO
from magma.t import Kind, In, Out
from magma.view import PortView


class _XMRPrimitiveBase(Generator2):
    def __init__(self, value: PortView):
        self.T = T = type(value)._to_magma_()
        self.io = io = IO(**type(self).make_ports(T))
        self.primitive = True

        self._value = value

    @property
    def value(self) -> PortView:
        return self._value


class XMRSink(_XMRPrimitiveBase):
    @staticmethod
    def make_ports(T: Kind):
        return {"I": In(T)}


class XMRSource(_XMRPrimitiveBase):
    @staticmethod
    def make_ports(T: Kind):
        return {"O": Out(T)}



from magma.bit import Bit
from magma.t import Type
from magma.tuple import Product, ProductKind


class ValidKind(ProductKind):
    def __getitem__(cls, T: Type):
        if not issubclass(T, Type):
            raise TypeError(
                f"Valid[T] expected T to be a subclass of m.Type not {T}"
            )
        fields = {"valid": Bit, "data": T}
        return type(f"Valid[{T}]", (Valid, ), fields)


class Valid(Product, metaclass=ValidKind):
    pass



import abc
import collections
from typing import Optional


class ScopedNameGeneratorBase(abc.ABC):
    @abc.abstractmethod
    def __call__(self, name: Optional[str] = None, force: bool = False) -> str:
        raise NotImplementedError()


class ScopedNameGenerator(ScopedNameGeneratorBase):
    def __init__(self):
        self._indices = collections.Counter()

    def __call__(self, name: Optional[str] = None, force: bool = False) -> str:
        if name is None:
            name = ""
        return self._call_impl(name, force)

    def _call_impl(self, name: str, force: bool) -> str:
        index = self._indices[name]
        self._indices[name] += 1
        if force:
            assert index == 0
            return name
        return name + str(index)



import io
import sys


class PrinterBase:
    def __init__(self, tab: int = 4, sout: io.TextIOBase = sys.stdout):
        self._tab = tab
        self._indent = 0
        self._sout = sout
        self._flushed = True

    def push(self):
        self._indent += 1

    def pop(self):
        if self._indent == 0:
            raise RuntimeError("Can not match deindent")
        self._indent -= 1

    def _make_indent(self) -> str:
        return f"{' ' * (self._indent * self._tab)}"

    def flush(self):
        self._sout.write("\n")
        self._flushed = True

    def print(self, s: str):
        tab = self._make_indent() if self._flushed else ""
        self._sout.write(f"{tab}{s}")
        self._flushed = False

    def print_line(self, line: str):
        self._sout.write(f"{self._make_indent()}{line}\n")
        self._flusehd = True



import contextlib
import itertools
from typing import Dict, Optional, Tuple, Union

from magma.bits import BitsMeta
from magma.clock import ClockTypes
from magma.circuit import CircuitKind, AnonymousCircuitType, CircuitBuilder
from magma.digital import DigitalMeta
from magma.generator import Generator2
from magma.interface import IO
from magma.logging import root_logger
from magma.passes.group import GrouperBase, InstanceCollection
from magma.primitives.mux import infer_mux_type
from magma.t import Kind, Type, In, Out
from magma.type_utils import type_to_sanitized_string


_logger = root_logger().getChild("compile_guard")


class _Grouper(GrouperBase):
    def __init__(self, instances: InstanceCollection, builder: CircuitBuilder):
        super().__init__(instances)
        self._builder = builder
        self._port_index = 0
        self._clock_types = set()

    def _visit_input_connection(self, _: Type, drivee: Type):
        # NOTE(rsetaluri): Since the driver might be traced (i.e. not an
        # immediate driver), we disregard it and grab the immediate driver
        # (value() vs. trace()).
        driver = drivee.value()
        new_port = self._builder.add_port(In(type(driver)))
        drivee.rewire(new_port)
        external = getattr(self._builder, new_port.name.name)
        external @= driver

    def _visit_output_connection(self, driver: Type, drivee: Type):
        new_port = self._builder.add_port(Out(type(driver)))
        new_port @= driver
        external = getattr(self._builder, new_port.name.name)
        drivee.rewire(external)

    def _visit_undriven_port(self, port: Type):
        # For undriven clock types, we simply lift the port *but do not connect
        # it* since we expect auto-wiring to do this for us (in fact, this is
        # why the port is undriven in the first place). For non-clock types, we
        # simply log a debug message and allow the downstream circuit pipeline
        # to handle the undriven port.
        if not isinstance(port, ClockTypes):
            _logger.debug(f"found undriven port: {port}")
            return
        T = type(port).undirected_t
        if T in self._clock_types:
            return  # only add at most one port for each clock type
        self._clock_types.add(T)
        name = str(port.name)
        self._builder.add_port(In(T), name=name)


class _CompileGuardBuilder(CircuitBuilder):
    __default_defn_name_counter = itertools.count()

    @staticmethod
    def _make_default_defn_name() -> str:
        counter = next(_CompileGuardBuilder.__default_defn_name_counter)
        return f"CompileGuardCircuit_{counter}"

    def __init__(self, name: Optional[str], cond: str, type: str):
        if name is None:
            name = _CompileGuardBuilder._make_default_defn_name()
        super().__init__(name)
        self._cond = cond
        self._system_types_added = set()
        if type not in {"defined", "undefined"}:
            raise ValueError(f"Unexpected compile guard type: {type}")
        metadata = {"condition_str": cond, "type": type}
        self._set_inst_attr("coreir_metadata", {"compile_guard": metadata})
        self._set_definition_attr("_compile_guard_", metadata)
        self._num_ports = itertools.count()

    def add_port(self, T: Kind, name: Optional[str] = None) -> Type:
        if name is None:
            name = self._new_port_name()
        return self._add_port(name, T)

    def instances(self) -> InstanceCollection:
        return self._instances.copy()

    def open(self):
        return self._open()

    def _new_port_name(self) -> str:
        return f"port_{next(self._num_ports)}"


def get_compile_guard_data(
        inst_or_defn: Union[AnonymousCircuitType, CircuitKind]
) -> Optional[Dict]:
    """Returns compile guard metadata (compile guard condition string and type
    (ifdef or ifndef)) attached to an instance or definition, or None if it is
    not a compile guarded module.
    """
    if isinstance(inst_or_defn, AnonymousCircuitType):
        coreir_metadata = getattr(inst_or_defn, "coreir_metadata", {})
        return coreir_metadata.get("compile_guard", None)
    if isinstance(inst_or_defn, CircuitKind):
        return getattr(inst_or_defn, "_compile_guard_", None)
    raise TypeError(inst_or_defn)


def _make_builder(
        cond: str,
        defn_name: Optional[str],
        inst_name: Optional[str],
        type: str
) -> _CompileGuardBuilder:
    builder = _CompileGuardBuilder(defn_name, cond, type)
    if inst_name is not None:
        builder.set_instance_name(inst_name)
    return builder


@contextlib.contextmanager
def compile_guard(
        cond: str,
        defn_name: Optional[str] = None,
        inst_name: Optional[str] = None,
        type: str = "defined",
):
    builder = _make_builder(cond, defn_name, inst_name, type)
    with builder.open() as f:
        yield f
    grouper = _Grouper(builder.instances(), builder)
    grouper.run()


def _is_simple_type(T: Kind) -> bool:
    return isinstance(T, (DigitalMeta, BitsMeta))


class _CompileGuardSelect(Generator2):
    def __init__(self, T: Kind, keys: Tuple[str]):
        # NOTE(rsetaluri): We need to add this check because the implementation
        # of this generator emits verilog directly, and thereby requires that no
        # transformations happen to the port names/types. If the type is not
        # "simple" (i.e. Bit or Bits[N]) then the assumption breaks down and
        # this implementation will not work.
        if not _is_simple_type(T):
            raise TypeError(f"Unsupported type: {T}")
        num_keys = len(keys)
        assert num_keys > 1
        self.io = IO(**{f"I{i}": In(T) for i in range(num_keys)}, O=Out(T))
        self.verilog = ""
        for i, key in enumerate(keys):
            if i == 0:
                stmt = f"`ifdef {key}"
            elif key == "default":
                assert i == (num_keys - 1)
                stmt = "`else"
            else:
                stmt = f"`elsif {key}"
            self.verilog += f"""\
{stmt}
    assign O = I{i};
"""
        self.verilog += "`endif"
        T_str = type_to_sanitized_string(T)
        self.name = f"CompileGuardSelect_{T_str}_{'_'.join(keys)}"


def compile_guard_select(**kwargs):
    try:
        default = kwargs.pop("default")
    except KeyError:
        raise ValueError("Expected default argument") from None
    if not (len(kwargs) > 1):
        raise ValueError("Expected at least one key besides default")
    # We rely on insertion order to make the default the last element for the
    # generated if/elif/else code.
    kwargs["default"] = default
    T, _ = infer_mux_type(list(kwargs.values()))
    Select = _CompileGuardSelect(T, tuple(kwargs.keys()))
    return Select()(*kwargs.values())



import contextlib
import logging as py_logging
from typing import Any, List, Mapping
import weakref

from magma.common import Stack, Finalizable, FinalizableDelegator
from magma.logging import (
    stage_logger,
    unstage_logger,
    push_log_capturer,
    pop_log_capturer,
)
from magma.placer import PlacerBase


_VERILOG_FILE_OPEN = """
integer \\_file_{filename} ;
initial \\_file_{filename} = $fopen(\"{filename}\", \"{mode}\");
"""

_VERILOG_FILE_CLOSE = """
final $fclose(\\_file_{filename} );
"""

_DEFAULT_VERILOG_LOG_STR = """
`ifndef MAGMA_LOG_LEVEL
    `define MAGMA_LOG_LEVEL 1
`endif"""


def _inline_verilog(
        context: 'DefinitionContext',
        format_str: str,
        format_args: Mapping,
        symbol_table: Mapping,
        inline_wire_prefix: str = "_magma_inline_wire"):
    # NOTE(rsetaluri): These are hacks to avoid a circular dependency.
    from magma.inline_verilog import inline_verilog_impl
    with definition_context_manager(context):
        inline_verilog_impl(
            format_str, format_args, symbol_table, inline_wire_prefix)


class VerilogDisplayManager(Finalizable):
    def __init__(self, context: weakref.ReferenceType):
        self._context = context
        self._files = []
        self._displays = []
        self._insert_default_log_level = False

    @property
    def context(self) -> 'DefinitionContext':
        return self._context()

    def add_file(self, file):
        self._files.append(file)

    def insert_default_log_level(self):
        self._insert_default_log_level = True

    def add_display(self, display):
        self._displays.append(display)

    def finalize(self):
        if self._insert_default_log_level:
            _inline_verilog(self.context, _DEFAULT_VERILOG_LOG_STR, {}, {})
        # Finalization needs to finalize (i) opens, (ii) displays, and (iii)
        # closes, in that order.
        for file in self._files:
            string = _VERILOG_FILE_OPEN.format(
                filename=file.filename, mode=file.mode)
            _inline_verilog(self.context, string, {}, {})
        for display in self._displays:
            _inline_verilog(self.context, *display.get_inline_verilog())
        for file in self._files:
            string = _VERILOG_FILE_CLOSE.format(filename=file.filename)
            _inline_verilog(self.context, string, {}, {})


_definition_context_stack = Stack()


def _get_definition_context_stack() -> Stack:
    global _definition_context_stack
    return _definition_context_stack


class DefinitionContext(FinalizableDelegator):
    def __init__(self, placer: PlacerBase):
        super().__init__()
        self._placer = placer
        self._builders = []
        self._logs = []
        self._metadata = {}
        self.add_child("display", VerilogDisplayManager(weakref.ref(self)))

    @property
    def placer(self) -> PlacerBase:
        return self._placer

    def add_builder(self, builder):
        self._builders.append(builder)

    @property
    def logs(self) -> List:
        return self._logs.copy()

    def add_log(self, log):
        self._logs.append(log)

    def add_logs(self, logs):
        self._logs.extend(logs)

    def get_metadata(self, key: str) -> Any:
        return self._metadata[key]

    def set_metadata(self, key: str, value: Any):
        self._metadata[key] = value

    def set_default_metadata(self, key: str, default: Any) -> Any:
        return self._metadata.setdefault(key, default)

    def place_instances(self, defn):
        self._placer = self._placer.finalize(defn)
        for builder in self._builders:
            inst = builder.finalize()
            self._placer.place(inst)

    def finalize(self, defn):
        super().finalize()


def push_definition_context(
        ctx: DefinitionContext, use_staged_logger: bool = False):
    push_log_capturer(ctx)
    if use_staged_logger:
        stage_logger()
    _get_definition_context_stack().push(ctx)


def pop_definition_context(
        use_staged_logger: bool = False) -> DefinitionContext:
    if use_staged_logger:
        unstage_logger()
    pop_log_capturer()
    return _get_definition_context_stack().pop()


def get_definition_context() -> DefinitionContext:
    return _get_definition_context_stack().peek()


@contextlib.contextmanager
def definition_context_manager(
        ctx: DefinitionContext, use_staged_logger: bool = False):
    push_definition_context(ctx, use_staged_logger)
    try:
        yield
    finally:
        if use_staged_logger:
            unstage_logger()
        popped_ctx = pop_definition_context(use_staged_logger)
        assert popped_ctx is ctx



import dataclasses
from typing import Any, ClassVar, List, Mapping, Optional, Tuple

from magma.backend.mlir.common import default_field
from magma.backend.mlir.mlir import (
    MlirDialect, MlirOp, MlirValue, MlirType, MlirSymbol, MlirAttribute,
    begin_dialect, end_dialect)
from magma.backend.mlir.mlir_printer_utils import (
    print_names, print_types, print_signature, print_attr_dict)
from magma.backend.mlir.printer_base import PrinterBase


hw = MlirDialect("hw")
begin_dialect(hw)


@dataclasses.dataclass(frozen=True)
class ArrayType(MlirType):
    dims: Tuple[int]
    T: MlirType

    def emit(self) -> str:
        dim_str = "x".join(map(str, self.dims))
        return f"!hw.array<{dim_str}x{self.T.emit()}>"


@dataclasses.dataclass(frozen=True)
class StructType(MlirType):
    fields: Tuple[Tuple[str, MlirType]]

    def emit(self) -> str:
        field_str = ", ".join(f"{k}: {t.emit()}" for k, t in self.fields)
        return f"!hw.struct<{field_str}>"


@dataclasses.dataclass(frozen=True)
class InOutType(MlirType):
    T: MlirType

    def emit(self) -> str:
        return f"!hw.inout<{self.T.emit()}>"


@dataclasses.dataclass(frozen=True)
class InnerRefAttr(MlirAttribute):
    module: MlirSymbol
    name: MlirSymbol

    def emit(self) -> str:
        return f"#hw.innerNameRef<{self.module.name}::{self.name.name}>"


@dataclasses.dataclass(frozen=True)
class ParamDeclAttr(MlirAttribute):
    name: str
    type: MlirType
    value: Optional[MlirAttribute] = None

    def emit(self) -> str:
        s = f"{self.name}: {self.type.emit()}"
        if self.value is None:
            return s
        return f"{s} = {self.value.emit()}"


@dataclasses.dataclass
class ModuleOpBase(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]
    name: MlirSymbol
    parameters: List[ParamDeclAttr] = default_field(list)

    def print_op(self, printer: PrinterBase):
        printer.print(f"hw.{self.op_name} {self.name.name}")
        if self.parameters:
            printer.print("<")
            printer.print(", ".join(param.emit() for param in self.parameters))
            printer.print(">")
        printer.print("(")
        print_signature(self.operands, printer)
        printer.print(") -> (")
        print_signature(self.results, printer, raw_names=True)
        printer.print(")")


@dataclasses.dataclass
class ModuleOp(ModuleOpBase):
    op_name: ClassVar[str] = "module"

    def __post_init__(self):
        self._block = self.new_region().new_block()

    def add_operation(self, operation: MlirOp):
        self._block.add_operation(operation)


@dataclasses.dataclass
class ModuleExternOp(ModuleOpBase):
    op_name: ClassVar[str] = "module.extern"


@dataclasses.dataclass
class OutputOp(MlirOp):
    operands: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        printer.print(f"hw.output ")
        print_names(self.operands, printer)
        printer.print(" : ")
        print_types(self.operands, printer)


@dataclasses.dataclass
class ConstantOp(MlirOp):
    results: List[MlirValue]
    value: int

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(f" = hw.constant {self.value} : ")
        print_types(self.results, printer)


@dataclasses.dataclass
class InstanceOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]
    name: str
    module: ModuleOpBase
    parameters: List[ParamDeclAttr] = default_field(list)
    sym: Optional[MlirSymbol] = None

    def print_op(self, printer: PrinterBase):
        if self.results:
            print_names(self.results, printer)
            printer.print(" = ")
        printer.print(f"hw.instance \"{self.name}\" ")
        if self.sym is not None:
            printer.print(f"sym {self.sym.name} ")
        printer.print(f"{self.module.name.name}")
        if self.parameters:
            printer.print("<")
            printer.print(", ".join(param.emit() for param in self.parameters))
            printer.print(">")
        printer.print("(")
        operands = [
            f"{m_operand.raw_name}: {operand.name}: {operand.type.emit()}"
            for operand, m_operand in zip(self.operands, self.module.operands)
        ]
        printer.print(", ".join(operands))
        printer.print(") -> (")
        results = [
            f"{m_result.raw_name}: {result.type.emit()}"
            for result, m_result in zip(self.results, self.module.results)
        ]
        printer.print(", ".join(results))
        printer.print(")")
        if self.attr_dict:
            printer.print(" ")
            print_attr_dict(self.attr_dict, printer)


@dataclasses.dataclass
class ArrayGetOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(" = hw.array_get ")
        print_names(self.operands[0], printer)
        printer.print("[")
        print_names(self.operands[1], printer)
        printer.print("] : ")
        print_types(self.operands[0], printer)


@dataclasses.dataclass
class ArraySliceOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(" = hw.array_slice ")
        print_names(self.operands[0], printer)
        printer.print(" at ")
        print_names(self.operands[1], printer)
        printer.print(" : (")
        print_types(self.operands[0], printer)
        printer.print(") -> ")
        print_types(self.results[0], printer)


@dataclasses.dataclass
class ArrayCreateOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(" = hw.array_create ")
        print_names(self.operands, printer)
        printer.print(" : ")
        print_types(self.operands[0], printer)


@dataclasses.dataclass
class ArrayConcatOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(" = hw.array_concat ")
        print_names(self.operands, printer)
        printer.print(" : ")
        print_types(self.operands, printer)


@dataclasses.dataclass
class StructExtractOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]
    field: str

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(" = hw.struct_extract ")
        print_names(self.operands, printer)
        printer.print(f"[\"{self.field}\"] : ")
        print_types(self.operands, printer)


@dataclasses.dataclass
class StructCreateOp(MlirOp):
    operands: List[MlirValue]
    results: List[MlirValue]

    def print_op(self, printer: PrinterBase):
        print_names(self.results, printer)
        printer.print(" = hw.struct_create (")
        print_names(self.operands, printer)
        printer.print(") : ")
        print_types(self.results, printer)


end_dialect()



from abc import abstractmethod, ABCMeta
import collections
import functools
import weakref
from .circuit import (DefineCircuitKind, Circuit, DebugCircuit,
                      DebugDefineCircuitKind)
from . import cache_definition
from magma.common import ParamDict
from hwtypes import BitVector


class GeneratorMeta(type):
    def __new__(mcs, name, bases, attrs):
        attrs["bind_generators"] = []
        cls = super().__new__(mcs, name, bases, attrs)
        old_generate = cls.generate

        def generate_wrapper(*args, **kwargs):
            result = old_generate(*args, **kwargs)
            if hasattr(result, "circuit_definition"):
                result = result.circuit_definition
            for gen in cls.bind_generators:
                gen.generate_bind(result, *args, **kwargs)
            return result

        cls.generate = generate_wrapper
        # Cache after bind logic so we don't run it twice on a cached entry
        if cls.cache:
            cls.generate = cache_definition(cls.generate)
        return cls

    def __call__(cls, *args, name=None, **kwargs):
        """
        name is reserved kwarg for naming instances

        TODO: Generalize handling of instance parameters
        """
        inst_kwargs = {}
        if name is not None:
            inst_kwargs["name"] = name
        return cls.generate(*args, **kwargs)(**inst_kwargs)


class Generator(metaclass=GeneratorMeta):
    # User can disable cacheing by setting this attribute to False
    cache = True

    @staticmethod
    @abstractmethod
    def generate(*args, **kwargs):
        raise NotImplementedError()

    @classmethod
    def bind(cls, monitor):
        cls.bind_generators.append(monitor)


def _make_key(cls, *args, **kwargs):
    _SECRET_KEY = "__magma_generator2_secret_key__"
    dct = {f"{_SECRET_KEY}{i}": v for i, v in enumerate(args)}
    dct.update(kwargs)
    for k, v in dct.items():
        if isinstance(v, BitVector):
            # Custom hash for BitVector to avoid inconsistent size error in key
            # comparison
            # TODO(leonardt/array2): We could move this to the BV.__hash__ if
            # this might be more generally useful as a pattern
            dct[k] = (int(v), len(v))
    return (cls, ParamDict(dct))


def _make_type(cls, *args, **kwargs):
    dummy = type.__new__(cls, "", (), {})
    name = cls.__name__
    bases = (cls._base_cls_,)
    dct = cls._base_metacls_.__prepare__(name, bases)
    cls.__init__(dummy, *args, **kwargs)
    dct.update(dict(dummy.__dict__))
    # NOTE(leonardt): We need to override the Generator2 classmethod bind with
    # DefineCircuitKind.bind for generator instances (circuits).
    dct["bind"] = classmethod(cls._base_metacls_.bind)
    ckt = cls._base_metacls_.__new__(cls, name, bases, dct)
    ckt._args_ = args
    ckt._kwargs_ = kwargs
    for gen in cls.bind_generators:
        gen.generate_bind(ckt, *args, **kwargs)
    return ckt


class _Generator2Meta(type):
    _cache = weakref.WeakValueDictionary()
    _base_cls_ = Circuit
    _base_metacls_ = DefineCircuitKind

    def __new__(metacls, name, bases, dct):
        bases = bases + (metacls._base_metacls_,)
        assert dct.setdefault("bind_generators", []) == []
        return type.__new__(metacls, name, bases, dct)

    def __call__(cls, *args, **kwargs):
        if cls is Generator2:
            try:
                name, bases, dct = args
            except ValueError:
                raise Exception("Can not initialize base Generator class")
            assert not kwargs
            return type.__new__(cls, name, bases, dct)
        if not getattr(cls, "_cache_", True):
            return _make_type(cls, *args, **kwargs)
        key = _make_key(cls, *args, **kwargs)
        try:
            return _Generator2Meta._cache[key]
        except KeyError:
            pass
        this = _make_type(cls, *args, **kwargs)
        _Generator2Meta._cache[key] = this
        return this


Generator2Kind = _Generator2Meta


class Generator2(metaclass=_Generator2Meta):
    def __new__(metacls, name, bases, dct):
        return type.__new__(metacls, name, bases, dct)

    def __call__(cls, *args, **kwargs):
        return type(cls)._base_metacls_.__call__(cls, *args, **kwargs)

    @classmethod
    def bind(cls, monitor):
        cls.bind_generators.append(monitor)


class _DebugGeneratorMeta(_Generator2Meta):
    _base_cls_ = DebugCircuit
    _base_metacls_ = DebugDefineCircuitKind


class DebugGenerator2(Generator2, metaclass=_DebugGeneratorMeta):
    pass


def reset_generator_cache():
    _Generator2Meta._cache = weakref.WeakValueDictionary()



import abc
import contextlib
import dataclasses
from typing import List, Mapping, Optional, Tuple
import weakref

from magma.backend.mlir.common import WithId, default_field, constant
from magma.backend.mlir.printer_base import PrinterBase
from magma.common import Stack


OptionalWeakRef = Optional[weakref.ReferenceType]


class DialectKind(abc.ABCMeta):
    def __new__(metacls, name, bases, dct):
        cls = super().__new__(metacls, name, bases, dct)
        try:
            maybe_peek_dialect
        except NameError:
            return cls
        dialect = maybe_peek_dialect()
        if dialect is not None:
            cls._register_(dialect)
        return cls


class MlirTypeMeta(DialectKind):
    def _register_(cls, dialect):
        dialect.register_type(cls)


class MlirType(metaclass=MlirTypeMeta):
    @abc.abstractmethod
    def emit(self) -> str:
        raise NotImplementedError()


@dataclasses.dataclass(frozen=True)
class MlirValue:
    type: MlirType
    raw_name: str

    @property
    def name(self) -> str:
        return f"%{self.raw_name}"


@dataclasses.dataclass(frozen=True)
class MlirSymbol:
    raw_name: str

    @property
    def name(self) -> str:
        return f"@{self.raw_name}"


class MlirAttributeMeta(DialectKind):
    def _register_(cls, dialect):
        dialect.register_attribute(cls)


class MlirAttribute(metaclass=MlirAttributeMeta):
    @abc.abstractmethod
    def emit(self) -> str:
        raise NotImplementedError()


@dataclasses.dataclass
class MlirBlock(WithId):
    operations: List['MlirOp'] = default_field(list, init=False)
    arguments: List[MlirValue] = default_field(list, init=False)
    # parent is of type MlirRegion.
    parent: OptionalWeakRef = default_field(constant(None), init=False)

    def add_operation(self, operation: 'MlirOp'):
        operation.set_parent(self)
        self.operations.append(operation)

    def set_parent(self, parent: 'MlirRegion'):
        self.parent = weakref.ref(parent)

    def print(self, printer: PrinterBase):
        for operation in self.operations:
            operation.print(printer)


_block_stack = Stack()


def get_block_stack() -> Stack:
    global _block_stack
    return _block_stack


@contextlib.contextmanager
def push_block(block):
    get_block_stack().push(block)
    try:
        yield
    finally:
        get_block_stack().pop()


@dataclasses.dataclass
class MlirRegion(WithId):
    blocks: List[MlirBlock] = default_field(list, init=False)
    # @parent is of type MlirOp.
    parent: OptionalWeakRef = default_field(constant(None), init=False)

    def new_block(self) -> MlirBlock:
        block = MlirBlock()
        block.set_parent(self)
        self.blocks.append(block)
        return block

    def set_parent(self, parent: 'MlirOp'):
        self.parent = weakref.ref(parent)

    def print(self, printer: PrinterBase):
        for block in self.blocks:
            block.print(printer)


class MlirOpMeta(DialectKind):
    def _register_(cls, dialect):
        dialect.register_op(cls)

    def __call__(cls, *args, **kwargs):
        obj = super().__call__(*args, **kwargs)
        try:
            block = get_block_stack().peek()
        except IndexError:
            pass
        else:
            block.add_operation(obj)
        return obj


@dataclasses.dataclass
class MlirOp(WithId, metaclass=MlirOpMeta):
    regions: List[MlirRegion] = default_field(list, init=False)
    operands: List[MlirValue] = default_field(list, init=False)
    results: List[MlirValue] = default_field(list, init=False)
    attr_dict: Mapping = default_field(dict, init=False)
    # @parent is of type MlirBlock.
    parent: OptionalWeakRef = default_field(constant(None), init=False)

    def new_region(self) -> MlirRegion:
        region = MlirRegion()
        region.set_parent(self)
        self.regions.append(region)
        return region

    def set_parent(self, parent: MlirBlock):
        self.parent = weakref.ref(parent)

    def print(self, printer: PrinterBase):
        self.print_op(printer)
        if not self.regions:
            printer.flush()
            return
        printer.print(" {")
        printer.flush()
        printer.push()
        for region in self.regions:
            region.print(printer)
        printer.pop()
        printer.print_line("}")

    @abc.abstractmethod
    def print_op(self, printer: PrinterBase):
        raise NotImplementedError()


class MlirDialect:
    def __init__(self, name: str):
        self._name = name
        self._klasses = {}

    @property
    def name(self) -> str:
        return self._name

    def _register(self, cls: type, arg_name: str, base: type):
        if not issubclass(cls, base):
            raise TypeError(f"{arg_name} must be subclass of {base.__name__}")
        name = cls.__name__
        if name in self._klasses:
            raise RuntimeError(f"{name} already part of {self.name} dialect")
        self._klasses[name] = cls
        setattr(self, name, cls)

    def register_op(self, op: type):
        self._register(op, "op", MlirOp)

    def register_type(self, t: type):
        self._register(t, "t", MlirType)

    def register_attribute(self, attr: type):
        self._register(attr, "attr", MlirAttribute)


_dialect_stack = Stack()


def get_dialect_stack() -> Stack:
    global _dialect_stack
    return _dialect_stack


def begin_dialect(dialect: MlirDialect):
    get_dialect_stack().push(dialect)


def end_dialect():
    get_dialect_stack().pop()


def maybe_peek_dialect() -> Optional[MlirDialect]:
    stack = get_dialect_stack()
    try:
        return stack.peek()
    except IndexError:
        return None



import inspect
import collections
import magma
from magma.config import get_debug_mode


debug_info = collections.namedtuple("debug_info", ["filename", "lineno", "module"])


def get_callee_frame_info():
    callee_frame = inspect.currentframe()
    # FIXME: Right now we assume a max 10 frames deep
    for i in range(0, 10):
        module = inspect.getmodule(callee_frame)
        # Go up until we're out of the magma module (assuming this is the user
        # code)
        if not module or module.__name__.split(".")[0] != "magma":
            break
        callee_frame = callee_frame.f_back
    callee_frame = inspect.getframeinfo(callee_frame)
    return debug_info(callee_frame.filename, callee_frame.lineno, module)


def debug_wire(fn):
    """
    Automatically populates the `debug_info` argument for a wire call if it's
    not already passed as an argument
    """
    # TODO: We could check that fn has the correct interface
    #       wire(i, o, debug_info)
    def wire(i, o=None, debug_info=None):
        if get_debug_mode() and debug_info is None:
            debug_info = get_callee_frame_info()
        return fn(i, o, debug_info)
    return wire

